from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import os
import time
import hydra
import lightning as L
#import rootutils
import pickle 
import torch
import numpy as np

from lightning import Callback, LightningDataModule, LightningModule, Trainer
from lightning.pytorch.loggers import Logger
from omegaconf import DictConfig, OmegaConf

#rootutils.setup_root(__file__, indicator=".project-root", pythonpath=True)

from pinnstorch import utils
from pinnstorch.data import (
    Interval,
    Mesh,
    PointCloud,
    Rectangle,
    RectangularPrism,
    TimeDomain,
)

from torch.profiler import profile, record_function, ProfilerActivity


log = utils.get_pylogger(__name__)

OmegaConf.register_new_resolver("eval", eval)

@utils.task_wrapper
def train(
    cfg: DictConfig, read_data_fn: Callable, pde_fn: Callable, output_fn: Callable = None
) -> Tuple[Dict[str, Any], Dict[str, Any]]:
    """Trains the model. Can additionally evaluate on a testset, using best weights obtained during
    training.

    This method is wrapped in optional @task_wrapper decorator, that controls the behavior during
    failure. Useful for multiruns, saving info about the crash, etc.

    :param cfg: A DictConfig configuration composed by Hydra.
    :return: A tuple with metrics and dict with all instantiated objects.
    """

    cfg = utils.set_mode(cfg)

    # set seed for random number generators in pytorch, numpy and python.random
    if cfg.get("seed"):
        L.seed_everything(cfg.seed, workers=True)
    
    if cfg.get("time_domain"):
        log.info(f"Instantiating time domain <{cfg.time_domain._target_}>")
        td: TimeDomain = hydra.utils.instantiate(cfg.time_domain)

    if cfg.get("spatial_domain"):
        log.info(f"Instantiating spatial domain <{cfg.spatial_domain._target_}>")
        sd: Union[Interval, Rectangle, RectangularPrism] = hydra.utils.instantiate(
            cfg.spatial_domain
        )

    log.info(f"Instantiating mesh <{cfg.mesh._target_}>")
    if cfg.mesh._target_ == "pinnstorch.data.Mesh":
        mesh: Mesh = hydra.utils.instantiate(
            cfg.mesh, time_domain=td, spatial_domain=sd, read_data_fn=read_data_fn
        )
    elif cfg.mesh._target_ == "pinnstorch.data.PointCloud":
        mesh: PointCloud = hydra.utils.instantiate(cfg.mesh, read_data_fn=read_data_fn)
    else:
        raise "Mesh should be defined in config file."

    train_datasets = []
    for i, (dataset_dic) in enumerate(cfg.train_datasets):
        for key, dataset in dataset_dic.items():
            log.info(f"Instantiating training dataset number {i+1}: <{dataset._target_}>")
            train_datasets.append(hydra.utils.instantiate(dataset)(mesh=mesh))

    val_dataset = None
    if cfg.get("val_dataset"):
        for i, dataset_dic in enumerate(cfg.val_dataset):
            for key, dataset in dataset_dic.items():
                log.info(f"Instantiating validation dataset number {i+1}: <{dataset._target_}>")
                val_dataset = hydra.utils.instantiate(dataset)(mesh=mesh)

    test_dataset = None
    if cfg.get("test_dataset"):
        for i, dataset_dic in enumerate(cfg.test_dataset):
            for key, dataset in dataset_dic.items():
                log.info(f"Instantiating test dataset number {i+1}: <{dataset._target_}>")
                test_dataset = hydra.utils.instantiate(dataset)(mesh=mesh)

    pred_dataset = None
    if cfg.get("pred_dataset"):
        for i, dataset_dic in enumerate(cfg.pred_dataset):
            for key, dataset in dataset_dic.items():
                log.info(f"Instantiating prediction dataset number {i+1}: <{dataset._target_}>")
                pred_dataset = hydra.utils.instantiate(dataset)(mesh=mesh)

    log.info(f"Instantiating datamodule <{cfg.data._target_}>")
    datamodule: LightningDataModule = hydra.utils.instantiate(
        cfg.data,
        train_datasets=train_datasets,
        val_dataset=val_dataset,
        test_dataset=test_dataset,
        pred_dataset=pred_dataset,
        batch_size=cfg.get("batch_size"),
    )

    if cfg.net._target_ == "pinnstorch.models.FCN":
        log.info(f"Instantiating neural net <{cfg.net._target_}>")
        net: torch.nn.Module = hydra.utils.instantiate(cfg.net)(lb=mesh.lb, ub=mesh.ub)
    elif cfg.net._target_ == "pinnstorch.models.NetHFM":
        # TODO
        log.info(f"Instantiating neural net <{cfg.net._target_}>")
        net: torch.nn.Module = hydra.utils.instantiate(cfg.net)(
            mean=train_datasets[0].mean, std=train_datasets[0].std
        )

    log.info(f"Instantiating model <{cfg.model._target_}>")

    model: LightningModule = hydra.utils.instantiate(cfg.model)(
        net=net, pde_fn=pde_fn, output_fn=output_fn
    )
    
    log.info("Instantiating callbacks...")
    callbacks: List[Callback] = utils.instantiate_callbacks(cfg.get("callbacks"))

    log.info("Instantiating loggers...")
    logger: List[Logger] = utils.instantiate_loggers(cfg.get("logger"))

    log.info(f"Instantiating trainer <{cfg.trainer._target_}>")

    trainer: Trainer = hydra.utils.instantiate(cfg.trainer, callbacks=callbacks, logger=logger)
    
    object_dict = {
        "cfg": cfg,
        "datamodule": datamodule,
        "model": model,
        "callbacks": callbacks,
        "logger": logger,
        "trainer": trainer,
    }

    if logger:
        log.info("Logging hyperparameters!")
        utils.log_hyperparameters(object_dict)

    if cfg.get("train"):
        log.info("Starting training!")
        start_time = time.time()
        trainer.fit(model=model, datamodule=datamodule)
        log.info(f"Elapsed time: {time.time() - start_time}")

    log.info(f"Median time for each batch: {np.median(model.times)}")
    
    train_metrics = trainer.callback_metrics
    
    if cfg.get("val"):
        log.info("Starting validation!")
        model.amp = False
        trainer.validate(model=model, datamodule=datamodule)

    if cfg.get("test"):
        log.info("Starting testing!")
        ckpt_path = None  # trainer.checkpoint_callback.best_model_path
        trainer.test(model=model, datamodule=datamodule, ckpt_path=ckpt_path)

    test_metrics = trainer.callback_metrics

    if cfg.get("pred_dataset"):
        preds_list = trainer.predict(
            model=model, datamodule=datamodule, ckpt_path=cfg.get("ckpt_path")
        )
        preds_dict = utils.fix_predictions(preds_list)
        if cfg.get("save_pred"):
            pred_path = f'{cfg.paths.output_dir}/predictions.pkl'
            with open(pred_path, 'wb') as f:
                pickle.dump(preds_dict, f)
            log.info(f"Predictions saved at: {pred_path}")
        
    if cfg.get("plotting"):
        log.info("Plotting the results")
        hydra.utils.instantiate(
            cfg.plotting,
            mesh=mesh,
            preds=preds_dict,
            train_datasets=train_datasets,
            val_dataset=val_dataset,
            file_name=cfg.paths.output_dir,
        )()

    # merge train and test metrics
    metric_dict = {**train_metrics, **test_metrics}

    return metric_dict, object_dict


@hydra.main(version_base="1.3", config_path="configs", config_name="train.yaml")
def main(cfg: DictConfig) -> Optional[float]:
    """Main entry point for training.

    :param cfg: DictConfig configuration composed by Hydra.
    :return: Optional[float] with optimized metric value.
    """
    # apply extra utilities
    # (e.g. ask for tags if none are provided in cfg, print cfg tree, etc.)
    utils.extras(cfg)

    # train the model
    metric_dict, _ = train(cfg)

    # safely retrieve metric value for hydra-based hyperparameter optimization
    metric_value = utils.get_metric_value(
        metric_dict=metric_dict, metric_name=cfg.get("optimized_metric")
    )

    # return optimized metric
    return metric_value


if __name__ == "__main__":
    main()



from typing import Tuple, List, Dict

import numpy as np
import torch

from pinnstorch.utils import jit_compiler

class SamplerBase:
    """Other samplers will be used this helper class."""

    def __init__(self):
        """Base class for sampling mesh data for training.

        Initializes instance variables for storing sampled data and solutions.
        """
        self.time_domain_sampled = None
        self.spatial_domain_sampled = None
        self.solution_sampled = None
        self.solution_names = None
        self.first_batch = True
        self.zeros_tensor = torch.tensor(0.0, dtype=torch.float32)
        

    def concatenate_solutions(self, flatten_mesh):
        """Concatenate dictionary of sampled solution data.

        :param flatten_mesh: Flattened mesh data.
        :return: Flattened mesh data with concatenated solutions.
        """
        flatten_mesh = list(flatten_mesh)
        concatenated_solutions = [
            flatten_mesh[2][solution_name] for solution_name in self.solution_names
        ]
        flatten_mesh[2] = np.concatenate(concatenated_solutions, axis=-1)

        return flatten_mesh

    def sample_mesh(self, num_sample, flatten_mesh):
        """Sample the mesh data for training.

        :param num_sample: Number of samples to generate.
        :param flatten_mesh: Flattened mesh data.
        :return: Sampled spatial, time, and solution data.
        """

        flatten_mesh = self.concatenate_solutions(flatten_mesh)

        if num_sample is None:
            return self.convert_to_tensor(flatten_mesh)
        else:
            idx = np.random.choice(range(flatten_mesh[0].shape[0]), num_sample, replace=False)
            return self.convert_to_tensor(
                (flatten_mesh[0][idx, :], flatten_mesh[1][idx, :], flatten_mesh[2][idx, :])
            )

    @staticmethod
    def convert_to_tensor(arrays):
        """Convert NumPy arrays to PyTorch tensors.

        :param arrays: List of NumPy arrays to convert.
        :return: List of converted PyTorch tensors.
        """

        return [torch.from_numpy(array.astype(np.float32)) for array in arrays]

    def loss_fn(self, inputs, loss, functions) -> Tuple[torch.Tensor, Dict[str, torch.Tensor]]:
        """Compute the loss function based on inputs and functions.

        :param inputs: Input data for computing the loss.
        :param loss: Loss variable.
        :param functions: Additional functions required for loss computation.
        :return: Loss variable and outputs dict from the forward pass.
        """
        
        self.functions = functions
        if self.first_batch and self.functions["jit_compile"] and not self.functions["val"]:
            if self.functions["batch_size"]:
                self.functions["pde_fn"] = torch.jit.script(self.functions["pde_fn"])
            else:
                self._loss_fn = torch.compile(self._loss_fn, backend=jit_compiler)
            self.first_batch = False

        loss, outputs = self._loss_fn(inputs, loss)

        return loss, outputs

    @property
    def mean(self):
        """Calculate the mean of the concatenated input data along each column.

        :return: A numpy array containing the mean values along each column.
        """

        x, t, _ = self[:]
        inputs = np.concatenate((*x, t), 1)

        return inputs.mean(0, keepdims=True)

    @property
    def std(self):
        """Calculate the standard deviation of the concatenated input data along each column.

        :return: A numpy array containing the standard deviation values along each column.
        """

        x, t, _ = self[:]
        inputs = np.concatenate((*x, t), 1)

        return inputs.std(0, keepdims=True)

    def __len__(self):
        """Get the number of sampled data points.

        :return: The number of sampled data points.
        """

        return len(self.spatial_domain_sampled[0])

    def __getitem__(self, idx):
        """Get a specific sampled data point using indexing. In some cases, we may not have
        `time_domain` and `solution_domain`. For example, in periodic boundary condition, there is
        not `solution_domain`.

        :param idx: Index of the desired data point.
        :return: Tuple containing spatial, time, and solution data for the indexed point.
        """

        spatial_domain = [spatial_domain[idx] for spatial_domain in self.spatial_domain_sampled]

        time_domain = self.zeros_tensor
        if self.time_domain_sampled is not None:
            time_domain = self.time_domain_sampled[idx]

        solution_domain = {'///': self.zeros_tensor}
        if self.solution_sampled is not None:
            solution_domain = {
                solution_name: self.solution_sampled[i][idx]
                for i, solution_name in enumerate(self.solution_names)
            }

        return (spatial_domain, time_domain, solution_domain)



import os
import logging
from itertools import combinations, product

import matplotlib as mpl
import matplotlib.gridspec as gridspec
import numpy as np
import scipy.io

from mpl_toolkits.axes_grid1 import make_axes_locatable
from mpl_toolkits.mplot3d import Axes3D
from scipy.interpolate import griddata

log = logging.getLogger(__name__)


def figsize(scale, nplots=1):
    """Calculate the figure size based on a given scale and number of plots.

    :param scale: Scaling factor for the figure size.
    :param nplots: Number of subplots in the figure (default is 1).
    :return: Calculated figure size in inches.
    """

    fig_width_pt = 390.0  # Get this from LaTeX using \the\textwidth
    inches_per_pt = 1.0 / 72.27  # Convert pt to inch
    golden_mean = (np.sqrt(5.0) - 1.0) / 2.0  # Aesthetic ratio (you could change this)
    fig_width = fig_width_pt * inches_per_pt * scale  # width in inches
    fig_height = nplots * fig_width * golden_mean  # height in inches
    fig_size = [fig_width, fig_height]
    return fig_size


mpl.rcParams.update(mpl.rcParamsDefault)

import matplotlib.pyplot as plt


def newfig(width, nplots=1):
    """Create a new figure with a specified width and number of subplots.

    :param width: Width of the figure.
    :param nplots: Number of subplots in the figure (default is 1).
    :return: Created figure and subplot axis.
    """

    fig = plt.figure(figsize=figsize(width, nplots))
    ax = fig.add_subplot(111)
    return fig, ax


def savefig(filename, crop=True):
    """Save a figure to the specified filename with optional cropping.

    :param filename: Name of the output file (without extension).
    :param crop: Whether to apply tight cropping to the saved image (default is True).
    """
    
    dir_name = os.path.dirname(filename)
    if dir_name and not os.path.exists(dir_name):
        os.makedirs(dir_name)
    

    if crop:
        plt.savefig(f"{filename}.pdf", bbox_inches="tight", pad_inches=0)
        plt.savefig(f"{filename}.eps", bbox_inches="tight", pad_inches=0)
    else:
        plt.savefig(f"{filename}.pdf")
        plt.savefig(f"{filename}.eps")

    log.info(f"Image saved at {filename}")


def plot_navier_stokes(mesh, preds, train_datasets, val_dataset, file_name):
    """Plot Navier-Stokes continuous inverse PDE."""

    x, t, u = train_datasets[0][:]
    p_star = mesh.solution["p"][:, 100]
    p_pred = preds["p"].reshape(p_star.shape)
    X_star = np.hstack(mesh.spatial_domain)
    lb = X_star.min(0)
    ub = X_star.max(0)
    nn = 200
    x = np.linspace(lb[0], ub[0], nn)
    y = np.linspace(lb[1], ub[1], nn)
    X, Y = np.meshgrid(x, y)
    x_star = X_star[:, 0:1]
    y_star = X_star[:, 1:2]

    PP_star = griddata(X_star, p_pred.flatten(), (X, Y), method="cubic")
    P_exact = griddata(X_star, p_star.flatten(), (X, Y), method="cubic")

    fig, ax = newfig(1.015, 0.8)
    ax.axis("off")
    gs2 = gridspec.GridSpec(1, 2)
    gs2.update(top=1, bottom=1 - 1 / 2, left=0.1, right=0.9, wspace=0.5)
    ax = plt.subplot(gs2[:, 0])
    h = ax.imshow(
        PP_star,
        interpolation="nearest",
        cmap="rainbow",
        extent=[x_star.min(), x_star.max(), y_star.min(), y_star.max()],
        origin="lower",
        aspect="auto",
    )
    divider = make_axes_locatable(ax)
    cax = divider.append_axes("right", size="5%", pad=0.05)

    fig.colorbar(h, cax=cax)
    ax.set_xlabel("$x$")
    ax.set_ylabel("$y$")
    ax.set_aspect("equal", "box")
    ax.set_title("Predicted pressure", fontsize=10)

    # Exact p(t,x,y)
    ax = plt.subplot(gs2[:, 1])
    h = ax.imshow(
        P_exact,
        interpolation="nearest",
        cmap="rainbow",
        extent=[x_star.min(), x_star.max(), y_star.min(), y_star.max()],
        origin="lower",
        aspect="auto",
    )
    divider = make_axes_locatable(ax)
    cax = divider.append_axes("right", size="5%", pad=0.05)

    fig.colorbar(h, cax=cax)
    ax.set_xlabel("$x$")
    ax.set_ylabel("$y$")
    ax.set_aspect("equal", "box")
    ax.set_title("Exact pressure", fontsize=10)
    savefig(file_name + "/fig")


def plot_kdv(mesh, preds, train_datasets, val_dataset, file_name):
    """Plot KdV discrete inverse PDE."""

    fig, ax = newfig(1.0, 1.2)
    ax.axis("off")

    x0 = train_datasets[0].spatial_domain_sampled[0].detach().numpy()
    u0 = train_datasets[0].solution_sampled[0].detach().numpy()
    idx_t0 = train_datasets[0].idx_t
    x1 = train_datasets[1].spatial_domain_sampled[0].detach().numpy()
    u1 = train_datasets[1].solution_sampled[0].detach().numpy()
    idx_t1 = train_datasets[1].idx_t
    exact_u = mesh.solution["u"]

    # Row 0: h(t,x)
    gs0 = gridspec.GridSpec(1, 2)
    gs0.update(top=1 - 0.06, bottom=1 - 1 / 2 + 0.1, left=0.15, right=0.85, wspace=0)
    ax = plt.subplot(gs0[:, :])

    h = ax.imshow(
        exact_u,
        interpolation="nearest",
        cmap="rainbow",
        extent=[
            mesh.time_domain[:].min(),
            mesh.time_domain[:].max(),
            mesh.spatial_domain[:].min(),
            mesh.spatial_domain[:].max(),
        ],
        origin="lower",
        aspect="auto",
    )
    divider = make_axes_locatable(ax)
    cax = divider.append_axes("right", size="5%", pad=0.05)
    fig.colorbar(h, cax=cax)

    line = np.linspace(mesh.spatial_domain[:].min(), mesh.spatial_domain[:].max(), 2)[:, None]
    ax.plot(mesh.time_domain[idx_t0] * np.ones((2, 1)), line, "w-", linewidth=1)
    ax.plot(mesh.time_domain[idx_t1] * np.ones((2, 1)), line, "w-", linewidth=1)

    ax.set_xlabel("$t$")
    ax.set_ylabel("$x$")
    leg = ax.legend(frameon=False, loc="best")
    ax.set_title("$u(t,x)$", fontsize=10)

    # Row 1: h(t,x) slices
    gs1 = gridspec.GridSpec(1, 2)
    gs1.update(top=1 - 1 / 2 - 0.05, bottom=0.15, left=0.15, right=0.85, wspace=0.5)

    ax = plt.subplot(gs1[0, 0])
    ax.plot(mesh.spatial_domain[:], exact_u[:, idx_t0].T, "b-", linewidth=2)
    ax.plot(x0, u0, "rx", linewidth=2, label="Data")
    ax.set_xlabel("$x$")
    ax.set_ylabel("$u(t,x)$")
    ax.set_title("$t = %.2f$" % (mesh.time_domain[idx_t0]), fontsize=10)
    ax.set_xlim([mesh.lb[:-1] - 0.1, mesh.ub[:-1] + 0.1])
    ax.legend(loc="upper center", bbox_to_anchor=(0.8, -0.3), ncol=2, frameon=False)

    ax = plt.subplot(gs1[0, 1])
    ax.plot(mesh.spatial_domain[:], exact_u[:, idx_t1], "b-", linewidth=2)
    ax.plot(x1, u1, "rx", linewidth=2, label="Data")
    ax.set_xlabel("$x$")
    ax.set_ylabel("$u(t,x)$")
    ax.set_title("$t = %.2f$" % (mesh.time_domain[idx_t1]), fontsize=10)
    ax.set_xlim([mesh.lb[:-1] - 0.1, mesh.ub[:-1] + 0.1])

    savefig(file_name + "/fig")


def plot_ac(mesh, preds, train_datasets, val_dataset, file_name):
    """Plot Allen-Cahn discrete forward PDE."""

    fig, ax = newfig(1.0, 1.2)

    x0 = train_datasets[0].spatial_domain_sampled[0].detach().numpy()
    u0 = train_datasets[0].solution_sampled[0].detach().numpy()
    exact_u = mesh.solution["u"]
    idx_t0 = train_datasets[0].idx_t
    idx_t1 = val_dataset.idx_t
    U1_pred = preds["u"]

    ax.axis("off")

    # Row 0: h(t,x)
    gs0 = gridspec.GridSpec(1, 2)
    gs0.update(top=1 - 0.06, bottom=1 - 1 / 2 + 0.1, left=0.15, right=0.85, wspace=0)
    ax = plt.subplot(gs0[:, :])

    h = ax.imshow(
        exact_u,
        interpolation="nearest",
        cmap="seismic",
        extent=[
            mesh.time_domain[:].min(),
            mesh.time_domain[:].max(),
            mesh.spatial_domain[:].min(),
            mesh.spatial_domain[:].max(),
        ],
        origin="lower",
        aspect="auto",
    )
    divider = make_axes_locatable(ax)
    cax = divider.append_axes("right", size="5%", pad=0.05)
    fig.colorbar(h, cax=cax)

    line = np.linspace(mesh.spatial_domain[:].min(), mesh.spatial_domain[:].max(), 2)[:, None]
    ax.plot(mesh.time_domain[idx_t0] * np.ones((2, 1)), line, "w-", linewidth=1)
    ax.plot(mesh.time_domain[idx_t1] * np.ones((2, 1)), line, "w-", linewidth=1)

    ax.set_xlabel("$t$")
    ax.set_ylabel("$x$")
    leg = ax.legend(frameon=False, loc="best")
    ax.set_title("$u(t,x)$", fontsize=10)

    # Row 1: h(t,x) slices
    gs1 = gridspec.GridSpec(1, 2)
    gs1.update(top=1 - 1 / 2 - 0.05, bottom=0.15, left=0.15, right=0.85, wspace=0.5)

    ax = plt.subplot(gs1[0, 0])
    ax.plot(mesh.spatial_domain[:], exact_u[:, idx_t0], "b-", linewidth=2)
    ax.plot(x0, u0, "rx", linewidth=2, label="Data")
    ax.set_xlabel("$x$")
    ax.set_ylabel("$u(t,x)$")
    ax.set_title("$t = %.2f$" % (mesh.time_domain[idx_t0]), fontsize=10)
    ax.set_xlim([mesh.lb[:-1] - 0.1, mesh.ub[:-1] + 0.1])
    ax.legend(loc="upper center", bbox_to_anchor=(0.8, -0.3), ncol=2, frameon=False)

    ax = plt.subplot(gs1[0, 1])
    ax.plot(mesh.spatial_domain[:], exact_u[:, idx_t1], "b-", linewidth=2, label="Exact")
    ax.plot(mesh.spatial_domain[:], U1_pred[:, -1], "r--", linewidth=2, label="Prediction")
    ax.set_xlabel("$x$")
    ax.set_ylabel("$u(t,x)$")
    ax.set_title("$t = %.2f$" % (mesh.time_domain[idx_t1]), fontsize=10)
    ax.set_xlim([mesh.lb[:-1] - 0.1, mesh.ub[:-1] + 0.1])

    ax.legend(loc="upper center", bbox_to_anchor=(0.1, -0.3), ncol=2, frameon=False)

    savefig(file_name + "/fig")


def plot_burgers_discrete_forward(mesh, preds, train_datasets, val_dataset, file_name):
    """Plot burgers discrete forward PDE."""

    fig, ax = newfig(1.0, 1.2)

    x0 = train_datasets[0].spatial_domain_sampled[0].detach().numpy()
    u0 = train_datasets[0].solution_sampled[0].detach().numpy()
    exact_u = mesh.solution["u"]
    idx_t0 = train_datasets[0].idx_t
    idx_t1 = val_dataset.idx_t
    U1_pred = preds["u"]

    ax.axis("off")

    # Row 0: h(t,x)
    gs0 = gridspec.GridSpec(1, 2)
    gs0.update(top=1 - 0.06, bottom=1 - 1 / 2 + 0.1, left=0.15, right=0.85, wspace=0)
    ax = plt.subplot(gs0[:, :])

    h = ax.imshow(
        exact_u,
        interpolation="nearest",
        cmap="seismic",
        extent=[
            mesh.time_domain[:].min(),
            mesh.time_domain[:].max(),
            mesh.spatial_domain[:].min(),
            mesh.spatial_domain[:].max(),
        ],
        origin="lower",
        aspect="auto",
    )
    divider = make_axes_locatable(ax)
    cax = divider.append_axes("right", size="5%", pad=0.05)
    fig.colorbar(h, cax=cax)

    line = np.linspace(mesh.spatial_domain[:].min(), mesh.spatial_domain[:].max(), 2)[:, None]
    ax.plot(mesh.time_domain[idx_t0] * np.ones((2, 1)), line, "w-", linewidth=1)
    ax.plot(mesh.time_domain[idx_t1] * np.ones((2, 1)), line, "w-", linewidth=1)

    ax.set_xlabel("$t$")
    ax.set_ylabel("$x$")
    leg = ax.legend(frameon=False, loc="best")
    ax.set_title("$u(t,x)$", fontsize=10)

    # Row 1: h(t,x) slices
    gs1 = gridspec.GridSpec(1, 2)
    gs1.update(top=1 - 1 / 2 - 0.05, bottom=0.15, left=0.15, right=0.85, wspace=0.5)

    ax = plt.subplot(gs1[0, 0])
    ax.plot(mesh.spatial_domain[:], exact_u[:, idx_t0], "b-", linewidth=2)
    ax.plot(x0, u0, "rx", linewidth=2, label="Data")
    ax.set_xlabel("$x$")
    ax.set_ylabel("$u(t,x)$")
    ax.set_title("$t = %.2f$" % (mesh.time_domain[idx_t0]), fontsize=10)
    ax.set_xlim([mesh.lb[:-1] - 0.1, mesh.ub[:-1] + 0.1])
    ax.legend(loc="upper center", bbox_to_anchor=(0.8, -0.3), ncol=2, frameon=False)

    ax = plt.subplot(gs1[0, 1])
    ax.plot(mesh.spatial_domain[:], exact_u[:, idx_t1], "b-", linewidth=2, label="Exact")
    ax.plot(mesh.spatial_domain[:], U1_pred[:, -1], "r--", linewidth=2, label="Prediction")
    ax.set_xlabel("$x$")
    ax.set_ylabel("$u(t,x)$")
    ax.set_title("$t = %.2f$" % (mesh.time_domain[idx_t1]), fontsize=10)
    ax.set_xlim([mesh.lb[:-1] - 0.1, mesh.ub[:-1] + 0.1])

    ax.legend(loc="upper center", bbox_to_anchor=(0.1, -0.3), ncol=2, frameon=False)

    savefig(file_name + "/fig")


def plot_burgers_discrete_inverse(mesh, preds, train_datasets, val_dataset, file_name):
    """Plot burgers continuous forward PDE."""

    fig, ax = newfig(1.0, 1.2)

    x0 = train_datasets[0].spatial_domain_sampled[0].detach().numpy()
    u0 = train_datasets[0].solution_sampled[0].detach().numpy()
    x1 = train_datasets[1].spatial_domain_sampled[0].detach().numpy()
    u1 = train_datasets[1].solution_sampled[0].detach().numpy()
    exact_u = mesh.solution["u"]
    idx_t0 = train_datasets[0].idx_t
    idx_t1 = val_dataset.idx_t

    ax.axis("off")

    fig, ax = newfig(1.0, 1.5)
    ax.axis("off")

    gs0 = gridspec.GridSpec(1, 2)
    gs0.update(top=1 - 0.06, bottom=1 - 1 / 3 + 0.05, left=0.15, right=0.85, wspace=0)
    ax = plt.subplot(gs0[:, :])

    h = ax.imshow(
        exact_u,
        interpolation="nearest",
        cmap="rainbow",
        extent=[
            mesh.time_domain[:].min(),
            mesh.time_domain[:].max(),
            mesh.spatial_domain[:].min(),
            mesh.spatial_domain[:].max(),
        ],
        origin="lower",
        aspect="auto",
    )
    divider = make_axes_locatable(ax)
    cax = divider.append_axes("right", size="5%", pad=0.05)
    fig.colorbar(h, cax=cax)

    line = np.linspace(mesh.spatial_domain[:].min(), mesh.spatial_domain[:].max(), 2)[:, None]
    ax.plot(mesh.time_domain[idx_t0] * np.ones((2, 1)), line, "w-", linewidth=1.0)
    ax.plot(mesh.time_domain[idx_t1] * np.ones((2, 1)), line, "w-", linewidth=1.0)
    ax.set_xlabel("$t$")
    ax.set_ylabel("$x$")
    ax.set_title("$u(t,x)$", fontsize=10)

    gs1 = gridspec.GridSpec(1, 2)
    gs1.update(top=1 - 1 / 3 - 0.1, bottom=1 - 2 / 3, left=0.15, right=0.85, wspace=0.5)

    ax = plt.subplot(gs1[0, 0])
    ax.plot(mesh.spatial_domain[:], exact_u[:, idx_t0][:, None], "b", linewidth=2, label="Exact")
    ax.plot(x0, u0, "rx", linewidth=2, label="Data")
    ax.set_xlabel("$x$")
    ax.set_ylabel("$u(t,x)$")
    ax.set_title(
        "$t = %.2f$\n%d training data" % (mesh.time_domain[idx_t0], u0.shape[0]), fontsize=10
    )

    ax = plt.subplot(gs1[0, 1])
    ax.plot(mesh.spatial_domain[:], exact_u[:, idx_t1][:, None], "b", linewidth=2, label="Exact")
    ax.plot(x1, u1, "rx", linewidth=2, label="Data")
    ax.set_xlabel("$x$")
    ax.set_ylabel("$u(t,x)$")
    ax.set_title(
        "$t = %.2f$\n%d training data" % (mesh.time_domain[idx_t1], u1.shape[0]), fontsize=10
    )
    ax.legend(loc="upper center", bbox_to_anchor=(-0.3, -0.3), ncol=2, frameon=False)

    savefig(file_name + "/fig")


def plot_schrodinger(mesh, preds, train_datasets, val_dataset, file_name):
    """Plot schrodinger continuous forward PDE."""

    h_pred = preds["h"]
    Exact_h = mesh.solution["h"]
    H_pred = h_pred.reshape(Exact_h.shape)

    # Row 1: u(t,x) slices
    gs1 = gridspec.GridSpec(1, 3)
    gs1.update(top=1 - 1 / 3, bottom=0, left=0.1, right=0.9, wspace=0.5)

    x0, t0, u0 = train_datasets[1][:]
    x_b, t_b, _ = train_datasets[2][:]
    mid = t_b.size()[0] // 2

    X0 = np.hstack((x0[0], t0))
    X_ub = np.hstack((x_b[0][:mid], t_b[:mid]))
    X_lb = np.hstack((x_b[0][mid:], t_b[mid:]))
    X_u_train = np.vstack([X0, X_lb, X_ub])

    fig, ax = newfig(1.0, 0.9)
    ax.axis("off")

    # Row 0: h(t,x)
    gs0 = gridspec.GridSpec(1, 2)
    gs0.update(top=1 - 0.06, bottom=1 - 1 / 3, left=0.15, right=0.85, wspace=0)
    ax = plt.subplot(gs0[:, :])

    h = ax.imshow(
        H_pred,
        interpolation="nearest",
        cmap="YlGnBu",
        extent=[mesh.lb[1], mesh.ub[1], mesh.lb[0], mesh.ub[0]],
        origin="lower",
        aspect="auto",
    )
    divider = make_axes_locatable(ax)
    cax = divider.append_axes("right", size="5%", pad=0.05)
    fig.colorbar(h, cax=cax)

    ax.plot(
        X_u_train[:, 1],
        X_u_train[:, 0],
        "kx",
        label="Data (%d points)" % (X_u_train.shape[0]),
        markersize=4,
        clip_on=False,
    )
    line = np.linspace(mesh.spatial_domain_mesh[:].min(), mesh.spatial_domain_mesh[:].max(), 2)[:, None]

    ax.plot(mesh.time_domain[75] * np.ones((2, 1)), line, "k--", linewidth=1)
    ax.plot(mesh.time_domain[100] * np.ones((2, 1)), line, "k--", linewidth=1)
    ax.plot(mesh.time_domain[125] * np.ones((2, 1)), line, "k--", linewidth=1)

    ax.set_xlabel("$t$")
    ax.set_ylabel("$x$")
    leg = ax.legend(frameon=False, loc="best")
    ax.set_title("$|h(t,x)|$", fontsize=10)

    # Row 1: h(t,x) slices
    gs1 = gridspec.GridSpec(1, 3)
    gs1.update(top=1 - 1 / 3, bottom=0, left=0.1, right=0.9, wspace=0.5)

    ax = plt.subplot(gs1[0, 0])
    ax.plot(mesh.spatial_domain_mesh[:, 75, 0], Exact_h[:, 75], "b-", linewidth=2, label="Exact")
    ax.plot(mesh.spatial_domain_mesh[:, 75, 0], H_pred[:, 75], "r--", linewidth=2, label="Prediction")
    ax.set_xlabel("$x$")
    ax.set_ylabel("$|h(t,x)|$")
    ax.set_title("$t = %.2f$" % (mesh.time_domain[75]), fontsize=10)
    ax.axis("square")
    ax.set_xlim([-5.1, 5.1])
    ax.set_ylim([-0.1, 5.1])

    ax = plt.subplot(gs1[0, 1])
    ax.plot(mesh.spatial_domain_mesh[:, 100, 0], Exact_h[:, 100], "b-", linewidth=2, label="Exact")
    ax.plot(mesh.spatial_domain_mesh[:, 100, 0], H_pred[:, 100], "r--", linewidth=2, label="Prediction")
    ax.set_xlabel("$x$")
    ax.set_ylabel("$|h(t,x)|$")
    ax.axis("square")
    ax.set_xlim([-5.1, 5.1])
    ax.set_ylim([-0.1, 5.1])
    ax.set_title("$t = %.2f$" % (mesh.time_domain[100]), fontsize=10)
    ax.legend(loc="upper center", bbox_to_anchor=(0.5, -0.8), ncol=5, frameon=False)

    ax = plt.subplot(gs1[0, 2])
    ax.plot(mesh.spatial_domain_mesh[:, 125, 0], Exact_h[:, 125], "b-", linewidth=2, label="Exact")
    ax.plot(mesh.spatial_domain_mesh[:, 125, 0], H_pred[:, 125], "r--", linewidth=2, label="Prediction")
    ax.set_xlabel("$x$")
    ax.set_ylabel("$|h(t,x)|$")
    ax.axis("square")
    ax.set_xlim([-5.1, 5.1])
    ax.set_ylim([-0.1, 5.1])
    ax.set_title("$t = %.2f$" % (mesh.time_domain[125]), fontsize=10)

    savefig(file_name + "/fig")


def plot_burgers_continuous_forward(mesh, preds, train_datasets, val_dataset, file_name):
    """Plot burgers continuous forward PDE."""

    U_pred = preds["u"]
    exact_u = mesh.solution["u"]
    x = mesh.spatial_domain[:]
    x_i, t_i, _ = train_datasets[1][:]
    x_b, t_b, _ = train_datasets[2][:]

    U_pred = U_pred.reshape(exact_u.shape)
    X_u_train = np.vstack([x_i[0], x_b[0]])

    X_u_time = np.vstack([t_i, t_b])

    X_u_train = np.hstack([X_u_train, X_u_time])
    fig, ax = newfig(1.5, 0.9)
    ax.axis("off")

    # Row 0: u(t,x)
    gs0 = gridspec.GridSpec(1, 2)
    gs0.update(top=1 - 0.06, bottom=1 - 1 / 3, left=0.15, right=0.85, wspace=0)
    ax = plt.subplot(gs0[:, :])

    h = ax.imshow(
        U_pred,
        interpolation="nearest",
        cmap="rainbow",
        extent=[
            mesh.time_domain[:].min(),
            mesh.time_domain[:].max(),
            mesh.spatial_domain[:].min(),
            mesh.spatial_domain[:].max(),
        ],
        origin="lower",
        aspect="auto",
    )
    divider = make_axes_locatable(ax)
    cax = divider.append_axes("right", size="5%", pad=0.05)
    fig.colorbar(h, cax=cax)

    ax.plot(
        X_u_train[:, 1],
        X_u_train[:, 0],
        "kx",
        label="Data (%d points)" % (X_u_train.shape[0]),
        markersize=4,
        clip_on=False,
    )

    line = np.linspace(mesh.spatial_domain[:].min(), mesh.spatial_domain[:].max(), 2)[:, None]
    ax.plot(mesh.time_domain[25] * np.ones((2, 1)), line, "w-", linewidth=1)
    ax.plot(mesh.time_domain[50] * np.ones((2, 1)), line, "w-", linewidth=1)
    ax.plot(mesh.time_domain[75] * np.ones((2, 1)), line, "w-", linewidth=1)

    ax.set_xlabel("$t$")
    ax.set_ylabel("$x$")
    ax.legend(frameon=False, loc="best")
    ax.set_title("$u(t,x)$", fontsize=10)

    # Row 1: u(t,x) slices
    gs1 = gridspec.GridSpec(1, 3)
    gs1.update(top=1 - 1 / 3, bottom=0, left=0.1, right=0.9, wspace=0.5)

    ax = plt.subplot(gs1[0, 0])
    ax.plot(x, exact_u[:, 25], "b-", linewidth=2, label="Exact")
    ax.plot(x, U_pred[:, 25], "r--", linewidth=2, label="Prediction")
    ax.set_xlabel("$x$")
    ax.set_ylabel("$u(t,x)$")
    ax.set_title("$t = 0.25$", fontsize=10)
    ax.axis("square")
    ax.set_xlim([-1.1, 1.1])
    ax.set_ylim([-1.1, 1.1])

    ax = plt.subplot(gs1[0, 1])
    ax.plot(x, exact_u[:, 50], "b-", linewidth=2, label="Exact")
    ax.plot(x, U_pred[:, 50], "r--", linewidth=2, label="Prediction")
    ax.set_xlabel("$x$")
    ax.set_ylabel("$u(t,x)$")
    ax.axis("square")
    ax.set_xlim([-1.1, 1.1])
    ax.set_ylim([-1.1, 1.1])
    ax.set_title("$t = 0.50$", fontsize=10)
    ax.legend(loc="upper center", bbox_to_anchor=(0.5, -0.35), ncol=5, frameon=False)

    ax = plt.subplot(gs1[0, 2])
    ax.plot(x, exact_u[:, 75], "b-", linewidth=2, label="Exact")
    ax.plot(x, U_pred[:, 75], "r--", linewidth=2, label="Prediction")
    ax.set_xlabel("$x$")
    ax.set_ylabel("$u(t,x)$")
    ax.axis("square")
    ax.set_xlim([-1.1, 1.1])
    ax.set_ylim([-1.1, 1.1])
    ax.set_title("$t = 0.75$", fontsize=10)

    savefig(file_name + "/fig")


def plot_burgers_continuous_inverse(mesh, preds, train_datasets, val_dataset, file_name):
    """Plot burgers continuous inverse PDE."""

    U_pred = preds["u"]

    exact_u = mesh.solution["u"]
    U_pred = U_pred.reshape(exact_u.shape)

    x_i, t_i, _ = train_datasets[0][:]

    X_u_train = np.hstack([x_i[0], t_i])

    fig, ax = newfig(1.0, 0.9)
    ax.axis("off")

    # Row 0: u(t,x)
    gs0 = gridspec.GridSpec(1, 2)
    gs0.update(top=1 - 0.06, bottom=1 - 1 / 3, left=0.15, right=0.85, wspace=0)
    ax = plt.subplot(gs0[:, :])

    h = ax.imshow(
        U_pred,
        interpolation="nearest",
        cmap="rainbow",
        extent=[
            mesh.time_domain[:].min(),
            mesh.time_domain[:].max(),
            mesh.spatial_domain[:].min(),
            mesh.spatial_domain[:].max(),
        ],
        origin="lower",
        aspect="auto",
    )
    divider = make_axes_locatable(ax)
    cax = divider.append_axes("right", size="5%", pad=0.05)
    fig.colorbar(h, cax=cax)

    ax.plot(
        X_u_train[:, 1],
        X_u_train[:, 0],
        "kx",
        label="Data (%d points)" % (X_u_train.shape[0]),
        markersize=4,
        clip_on=False,
    )

    line = np.linspace(mesh.spatial_domain[:].min(), mesh.spatial_domain[:].max(), 2)[:, None]
    ax.plot(mesh.time_domain[25] * np.ones((2, 1)), line, "w-", linewidth=1)
    ax.plot(mesh.time_domain[50] * np.ones((2, 1)), line, "w-", linewidth=1)
    ax.plot(mesh.time_domain[75] * np.ones((2, 1)), line, "w-", linewidth=1)

    ax.set_xlabel("$t$")
    ax.set_ylabel("$x$")
    ax.legend(frameon=False, loc="best")
    ax.set_title("$u(t,x)$", fontsize=10)

    # Row 1: u(t,x) slices
    gs1 = gridspec.GridSpec(1, 3)
    gs1.update(top=1 - 1 / 3, bottom=0, left=0.1, right=0.9, wspace=0.5)

    ax = plt.subplot(gs1[0, 0])
    ax.plot(mesh.spatial_domain[:], exact_u[:, 25], "b-", linewidth=2, label="Exact")
    ax.plot(mesh.spatial_domain[:], U_pred[:, 25], "r--", linewidth=2, label="Prediction")
    ax.set_xlabel("$x$")
    ax.set_ylabel("$u(t,x)$")
    ax.set_title("$t = 0.25$", fontsize=10)
    ax.axis("square")
    ax.set_xlim([-1.1, 1.1])
    ax.set_ylim([-1.1, 1.1])

    ax = plt.subplot(gs1[0, 1])
    ax.plot(mesh.spatial_domain[:], exact_u[:, 50], "b-", linewidth=2, label="Exact")
    ax.plot(mesh.spatial_domain[:], U_pred[:, 50], "r--", linewidth=2, label="Prediction")
    ax.set_xlabel("$x$")
    ax.set_ylabel("$u(t,x)$")
    ax.axis("square")
    ax.set_xlim([-1.1, 1.1])
    ax.set_ylim([-1.1, 1.1])
    ax.set_title("$t = 0.50$", fontsize=10)
    ax.legend(loc="upper center", bbox_to_anchor=(0.5, -0.35), ncol=5, frameon=False)

    ax = plt.subplot(gs1[0, 2])
    ax.plot(mesh.spatial_domain[:], exact_u[:, 75], "b-", linewidth=2, label="Exact")
    ax.plot(mesh.spatial_domain[:], U_pred[:, 75], "r--", linewidth=2, label="Prediction")
    ax.set_xlabel("$x$")
    ax.set_ylabel("$u(t,x)$")
    ax.axis("square")
    ax.set_xlim([-1.1, 1.1])
    ax.set_ylim([-1.1, 1.1])
    ax.set_title("$t = 0.75$", fontsize=10)

    savefig(file_name + "/fig")



from typing import Dict, List, Any, Tuple, Callable

import numpy as np
import torch
import torch._dynamo

from .sampler_base import SamplerBase
from pinnstorch.utils import set_requires_grad

class MeshSampler(SamplerBase):
    """Sample from Mesh for continuous mode."""

    def __init__(
        self,
        mesh,
        idx_t: int = None,
        num_sample: int = None,
        solution: List = None,
        collection_points: List = None,
        use_lhs: bool = True,
    ):
        """Initialize a mesh sampler for collecting training data.

        :param mesh: Instance of the mesh used for sampling.
        :param idx_t: Index of the time step.
        :param num_sample: Number of samples to generate.
        :param solution: Names of the solution outputs.
        :param collection_points: Names of the collection points.
        :param use_lhs: Whether use lhs or not for generating collection points.
        """

        super().__init__()

        self.solution_names = solution
        self.collection_points_names = collection_points
        self.idx_t = idx_t
        
        # On a time step.
        if self.idx_t:
            flatten_mesh = mesh.on_initial_boundary(self.solution_names, self.idx_t)

        # All time steps.
        elif self.solution_names is not None:
            flatten_mesh = mesh.flatten_mesh(self.solution_names)

        if self.solution_names:
            (
                self.spatial_domain_sampled,
                self.time_domain_sampled,
                self.solution_sampled,
            ) = self.sample_mesh(num_sample, flatten_mesh)

            self.solution_sampled = list(torch.split(self.solution_sampled, (1), dim=1))

        # Collection Points only.
        else:
            (self.spatial_domain_sampled, self.time_domain_sampled) = self.convert_to_tensor(
                mesh.collection_points(num_sample, use_lhs)
            )

            self.solution_sampled = None

        self.spatial_domain_sampled = list(torch.split(self.spatial_domain_sampled, (1), dim=1))
    
    def _loss_fn(self,
                 inputs,
                 loss) -> Tuple[torch.Tensor, Dict[str, torch.Tensor]]:
        """Compute the loss function based on inputs and functions.

        :param inputs: Input data for computing the loss.
        :param loss: Loss variable.
        :return: Loss variable and outputs dict from the forward pass.
        """
        x, t, u = inputs
        
        outputs = self.functions["forward"](x, t)

        if self.collection_points_names:
            if self.functions["extra_variables"] is not None:
                outputs = self.functions["pde_fn"](outputs, *x, t, self.functions["extra_variables"])
            else:    
                outputs = self.functions["pde_fn"](outputs, *x, t)

        loss = self.functions["loss_fn"](loss, outputs, keys=self.collection_points_names) + \
               self.functions["loss_fn"](loss, outputs, u, keys=self.solution_names)

        return loss, outputs
    
 
class DiscreteMeshSampler(SamplerBase):
    """Sample from Mesh for discrete mode."""

    def __init__(
        self,
        mesh,
        idx_t: int,
        num_sample: int = None,
        solution: List = None,
        collection_points: List = None,
    ):
        """Initialize a mesh sampler for collecting training data in discrete mode.

        :param mesh: Instance of the mesh used for sampling.
        :param idx_t: Index of the time step for discrete mode.
        :param num_sample: Number of samples to generate.
        :param solution: Names of the solution outputs.
        :param collection_points: Names of the collection points.
        """
        super().__init__()

        self.solution_names = solution
        self.collection_points_names = collection_points
        self.idx_t = idx_t
        self._mode = None

        flatten_mesh = mesh.on_initial_boundary(self.solution_names, self.idx_t)

        (
            self.spatial_domain_sampled,
            self.time_domain_sampled,
            self.solution_sampled,
        ) = self.sample_mesh(num_sample, flatten_mesh)

        self.spatial_domain_sampled = list(torch.split(self.spatial_domain_sampled, (1), dim=1))
        self.time_domain_sampled = None
        self.solution_sampled = list(torch.split(self.solution_sampled, (1), dim=1))

    @property
    def mode(self):
        """Get the current mode for RungeKutta class.

        :return: The current mode value.
        """
        return self._mode

    @mode.setter
    def mode(self, value):
        """Set the mode value by PINNDataModule for RungeKutta class.

        :param value: The mode value to be set.
        """
        self._mode = value

    def _loss_fn(self, inputs, loss):
        """Compute the loss function based on inputs and functions. _mode is assigned in
        PINNDataModule class. It can be `inverse_discrete_1`, `inverse_discrete_2`, or
        `forward_discrete`.

        :param inputs: Input data for computing the loss.
        :param loss: Loss variable.
        :return: Loss variable and outputs dict from the forward pass.
        """

        x, t, u = inputs

        outputs = self.functions["forward"](x, t)

        if self._mode:
            if self.functions["extra_variables"] is not None:
                outputs = self.functions["pde_fn"](outputs, *x, self.functions["extra_variables"])
            else:
                outputs = self.functions["pde_fn"](outputs, *x)
            outputs = self.functions["runge_kutta"](
                outputs,
                mode=self._mode,
                solution_names=self.solution_names,
                collection_points_names=self.collection_points_names,
            )
        loss = self.functions["loss_fn"](loss, outputs, u, keys=self.solution_names)

        return loss, outputs



import os
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple

import hydra
import numpy as np
import pytest
import rootutils
import torch
from hydra.core.hydra_config import HydraConfig
from omegaconf import DictConfig, open_dict

import pinnstorch
from pinnstorch.train import train
from tests.helpers.run_if import RunIf


def read_data_fn(root_path):
    """Read and preprocess data from the specified root path.

    :param root_path: The root directory containing the data.
    :return: Processed data will be used in Mesh class.
    """

    data = pinnstorch.utils.load_data(root_path, "KdV.mat")
    exact_u = np.real(data["uu"]).astype("float32")
    return {"u": exact_u}


def pde_fn(outputs, x, extra_variables):
    """Define the partial differential equations (PDEs)."""

    U = outputs["u"]
    U_x = pinnstorch.utils.fwd_gradient(U, x)[0]
    U_xx = pinnstorch.utils.fwd_gradient(U_x, x)[0]
    U_xxx = pinnstorch.utils.fwd_gradient(U_xx, x)[0]
    outputs["f"] = -extra_variables["l1"] * U * U_x - torch.exp(extra_variables["l2"]) * U_xxx
    return outputs


def test_train_config(cfg_train: DictConfig) -> None:
    """Tests the training configuration provided by the `cfg_train` pytest fixture.

    :param cfg_train: A DictConfig containing a valid training configuration.
    """
    assert cfg_train
    assert cfg_train.data
    assert cfg_train.model
    assert cfg_train.net
    assert cfg_train.trainer

    HydraConfig().set_config(cfg_train)

    net: torch.nn.Module = hydra.utils.instantiate(cfg_train.net)(lb=[0.0, 0.0], ub=[1.0, 1.0])
    hydra.utils.instantiate(cfg_train.model)(net=net, pde_fn=pde_fn, output_fn=None)
    hydra.utils.instantiate(cfg_train.trainer)


def test_eval_config(cfg_eval: DictConfig) -> None:
    """Tests the evaluation configuration provided by the `cfg_eval` pytest fixture.

    :param cfg_train: A DictConfig containing a valid evaluation configuration.
    """
    assert cfg_eval
    assert cfg_eval.data
    assert cfg_eval.model
    assert cfg_eval.trainer

    HydraConfig().set_config(cfg_eval)

    net: torch.nn.Module = hydra.utils.instantiate(cfg_eval.net)(lb=[0.0, 0.0], ub=[1.0, 1.0])
    hydra.utils.instantiate(cfg_eval.model)(net=net, pde_fn=pde_fn, output_fn=None)
    hydra.utils.instantiate(cfg_eval.trainer)



"""This file prepares config fixtures for other tests."""

from pathlib import Path

import pytest
import rootutils
from hydra import compose, initialize
from hydra.core.global_hydra import GlobalHydra
from omegaconf import DictConfig, open_dict


@pytest.fixture(scope="package")
def cfg_train_global() -> DictConfig:
    """A pytest fixture for setting up a default Hydra DictConfig for training.

    :return: A DictConfig object containing a default Hydra configuration for training.
    """
    with initialize(version_base="1.3", config_path="../examples/kdv/configs"):
        cfg = compose(config_name="config.yaml", return_hydra_config=True, overrides=[])

        # set defaults for all tests
        with open_dict(cfg):
            cfg.paths.root_dir = str(rootutils.find_root(indicator=".project-root"))
            cfg.trainer.max_epochs = 1
            cfg.trainer.limit_train_batches = 0.01
            cfg.trainer.limit_val_batches = 0.1
            cfg.trainer.limit_test_batches = 0.1
            cfg.trainer.accelerator = "cpu"
            cfg.trainer.devices = 1
            cfg.data.num_workers = 0
            cfg.data.pin_memory = False
            cfg.extras.print_config = False
            cfg.extras.enforce_tags = False
            cfg.logger = None

    return cfg


@pytest.fixture(scope="package")
def cfg_eval_global() -> DictConfig:
    """A pytest fixture for setting up a default Hydra DictConfig for evaluation.

    :return: A DictConfig containing a default Hydra configuration for evaluation.
    """
    with initialize(version_base="1.3", config_path="../examples/kdv/configs"):
        cfg = compose(config_name="config.yaml", return_hydra_config=True, overrides=[])

        # set defaults for all tests
        with open_dict(cfg):
            cfg.paths.root_dir = str(rootutils.find_root(indicator=".project-root"))
            cfg.trainer.max_epochs = 1
            cfg.trainer.limit_test_batches = 0.1
            cfg.trainer.accelerator = "cpu"
            cfg.trainer.devices = 1
            cfg.data.num_workers = 0
            cfg.data.pin_memory = False
            cfg.extras.print_config = False
            cfg.extras.enforce_tags = False
            cfg.logger = None

    return cfg


@pytest.fixture(scope="function")
def cfg_train(cfg_train_global: DictConfig, tmp_path: Path) -> DictConfig:
    """A pytest fixture built on top of the `cfg_train_global()` fixture, which accepts a temporary
    logging path `tmp_path` for generating a temporary logging path.

    This is called by each test which uses the `cfg_train` arg. Each test generates its own temporary logging path.

    :param cfg_train_global: The input DictConfig object to be modified.
    :param tmp_path: The temporary logging path.

    :return: A DictConfig with updated output and log directories corresponding to `tmp_path`.
    """
    cfg = cfg_train_global.copy()

    with open_dict(cfg):
        cfg.paths.output_dir = str(tmp_path)
        cfg.paths.log_dir = str(tmp_path)

    yield cfg

    GlobalHydra.instance().clear()


@pytest.fixture(scope="function")
def cfg_eval(cfg_eval_global: DictConfig, tmp_path: Path) -> DictConfig:
    """A pytest fixture built on top of the `cfg_eval_global()` fixture, which accepts a temporary
    logging path `tmp_path` for generating a temporary logging path.

    This is called by each test which uses the `cfg_eval` arg. Each test generates its own temporary logging path.

    :param cfg_train_global: The input DictConfig object to be modified.
    :param tmp_path: The temporary logging path.

    :return: A DictConfig with updated output and log directories corresponding to `tmp_path`.
    """
    cfg = cfg_eval_global.copy()

    with open_dict(cfg):
        cfg.paths.output_dir = str(tmp_path)
        cfg.paths.log_dir = str(tmp_path)

    yield cfg

    GlobalHydra.instance().clear()



from pinnstorch.utils.gradient_fn import fwd_gradient, gradient
from pinnstorch.utils.instantiators import instantiate_callbacks, instantiate_loggers
from pinnstorch.utils.logging_utils import log_hyperparameters
from pinnstorch.utils.module_fn import (
    fix_extra_variables,
    mse,
    relative_l2_error,
    sse,
    set_requires_grad,
    fix_predictions
)
from pinnstorch.utils.plotting import (
    plot_ac,
    plot_burgers_continuous_forward,
    plot_burgers_continuous_inverse,
    plot_burgers_discrete_forward,
    plot_burgers_discrete_inverse,
    plot_kdv,
    plot_navier_stokes,
    plot_schrodinger,
)
from pinnstorch.utils.pylogger import get_pylogger
from pinnstorch.utils.rich_utils import enforce_tags, print_config_tree
from pinnstorch.utils.utils import (
    extras,
    get_metric_value,
    load_data,
    load_data_txt,
    task_wrapper,
    set_mode
)

from pinnstorch.utils.jit_compiler import jit_compiler


from typing import Dict, List

import numpy as np
import torch
from torch import nn


class FCN(nn.Module):
    """A simple fully-connected neural net for solving equations.

    In this model, lower and upper bound will be used for normalization of input data
    """
    output_names: List[str]
    
    def __init__(self, layers, lb, ub, output_names, discrete: bool = False) -> None:
        """Initialize a `FCN` module.

        :param layers: The list indicating number of neurons in each layer.
        :param lb: Lower bound for the inputs.
        :param ub: Upper bound for the inputs.
        :param output_names: Names of outputs of net.
        :param discrete: If the problem is discrete or not.
        """
        super().__init__()

        self.model = self.initalize_net(layers)
        self.register_buffer("lb", torch.tensor(lb, dtype=torch.float32, requires_grad=False))
        self.register_buffer("ub", torch.tensor(ub, dtype=torch.float32, requires_grad=False))
        self.output_names = output_names
        self.discrete = discrete

    def initalize_net(self, layers: List):
        """Initialize the layers of the neural network.

        :param layers: The list indicating number of neurons in each layer.
        :return: The initialized neural network.
        """

        initializer = nn.init.xavier_uniform_
        net = nn.Sequential()

        input_layer = nn.Linear(layers[0], layers[1])
        initializer(input_layer.weight)

        net.add_module("input", input_layer)
        net.add_module("activation_1", nn.Tanh())

        for i in range(1, len(layers) - 2):
            hidden_layer = nn.Linear(layers[i], layers[i + 1])
            initializer(hidden_layer.weight)
            net.add_module(f"hidden_{i+1}", hidden_layer)
            net.add_module(f"activation_{i+1}", nn.Tanh())

        output_layer = nn.Linear(layers[-2], layers[-1])
        initializer(output_layer.weight)
        net.add_module("output", output_layer)
        return net

    def forward(self, spatial: List[torch.Tensor], time: torch.Tensor) -> Dict[str, torch.Tensor]:
        """Perform a single forward pass through the network.

        :param spatial: List of input spatial tensors.
        :param time: Input tensor representing time.
        :return: A tensor of solutions.
        """

        # Discrete Mode
        if self.discrete:
            if len(spatial) == 2:
                x, y = spatial
                z = torch.cat((x, y), 1)
            elif len(spatial) == 3:
                x, y, z = spatial
                z = torch.cat((x, y, z), 1)
            else:
                z = spatial[0]
            z = 2.0 * (z - self.lb[:-1]) / (self.ub[:-1] - self.lb[:-1]) - 1.0

        # Continuous Mode
        else:
            if len(spatial) == 1:
                x = spatial[0]
                z = torch.cat((x, time), 1)
            elif len(spatial) == 2:
                x, y = spatial
                z = torch.cat((x, y, time), 1)
            else:
                x, y, z = spatial
                z = torch.cat((x, y, z, time), 1)
            z = 2.0 * (z - self.lb) / (self.ub - self.lb) - 1.0

        z = self.model(z)

        # Discrete Mode
        if self.discrete:
            outputs_dict = {name: z for i, name in enumerate(self.output_names)}

        # Continuous Mode
        else:
            outputs_dict = {name: z[:, i : i + 1] for i, name in enumerate(self.output_names)}
        return outputs_dict


class NetHFM(nn.Module):
    """A simple fully-connected neural net for solving equations.

    In this model, mean and std will be used for normalization of input data. Also, weight
    normalization will be done.
    """
    output_names: List[str]
    
    def __init__(self, mean, std, layers: List, output_names: List):
        super().__init__()
        """Initialize a `NetHFM` module.

        :param mesh: The number of layers.
        :param layers: The list indicating number of neurons in each layer.
        :param output_names: Names of outputs of net.
        """
        self.num_layers = len(layers)
        self.output_names = output_names

        self.register_buffer("X_mean", torch.from_numpy(mean))
        self.register_buffer("X_std", torch.from_numpy(std))

        self.initalize_net(layers)
        self.sigmoid = nn.Sigmoid()

    def initalize_net(self, layers: List) -> None:
        """Initialize the neural network weights, biases, and gammas.

        :param layers: The list indicating number of neurons in each layer.
        """

        self.weights = []
        self.biases = []
        self.gammas = []

        for i in range(0, self.num_layers - 1):
            in_dim = layers[i]
            out_dim = layers[i + 1]
            W = np.random.normal(size=[in_dim, out_dim])
            b = np.zeros([1, out_dim])
            g = np.ones([1, out_dim])
            # tensorflow variables
            self.weights.append(torch.tensor(W, dtype=torch.float32))
            self.biases.append(torch.tensor(b, dtype=torch.float32))
            self.gammas.append(torch.tensor(g, dtype=torch.float32))

        self.weights = nn.ParameterList(self.weights)
        self.biases = nn.ParameterList(self.biases)
        self.gammas = nn.ParameterList(self.gammas)

    def forward(self, spatial: List[torch.Tensor], time: torch.Tensor) -> Dict[str, torch.Tensor]:
        """Perform a forward pass through the network.

        :param spatial: List of input spatial tensors.
        :param time: Input tensor representing time.
        :return: A dictionary with output names as keys and corresponding output tensors as values.
        """
        if len(spatial) == 1:
            x = spatial[0]
            H = torch.cat((x, time), 1)
        elif len(spatial) == 2:
            x, y = spatial
            H = torch.cat((x, y, time), 1)
        else:
            x, y, z = spatial
            H = torch.cat((x, y, z, time), 1)
        
        H = (H - self.X_mean) / self.X_std

        for i, (W, b, g) in enumerate(zip(self.weights, self.biases, self.gammas)):
            # weight normalization
            V = W / torch.norm(W, dim=0)
            # matrix multiplication
            H = torch.matmul(H, V)
            # add bias
            H = g * H + b
            # activation
            if i < self.num_layers - 2:
                H = H * self.sigmoid(H)
        
        outputs_dict = {name: H[:, i : i + 1] for i, name in enumerate(self.output_names)}

        return outputs_dict


if __name__ == "__main__":
    _ = FCN()
    _ = NetHFM()



import numpy as np


class TimeDomain:
    """Initialize a time domain."""

    def __init__(self, t_interval, t_points: int):
        """Initialize a TimeDomain object to represent a time domain.

        :param t_interval: A tuple or list representing the time interval [start_time, end_time].
        :param t_points: The number of time points to discretize the interval.
        """
        self.time_interval = t_interval
        self.time = np.linspace(self.time_interval[0], self.time_interval[1], num=t_points)

    def generate_mesh(self, spatial_points):
        """Generate a mesh of time based on number of spatial points.

        :param spatial_points: The number of spatial points.
        :return: A mesh of time and spatial points.
        """

        mesh = np.tile(self.time, (spatial_points, 1))[:, :, None]

        return mesh

    def __len__(self):
        """Get the length of the time domain.

        :return: The number of time points in the time domain.
        """
        return len(self.time)

    def __getitem__(self, idx):
        """Get a specific time point from the time domain using indexing.

        :param idx: The index of the desired time point.
        :return: The time value at the specified index.
        """
        return self.time[idx]



import os
import warnings
from importlib.util import find_spec
from typing import Any, Callable, Dict, Tuple

import numpy as np
import requests
import scipy
from omegaconf import DictConfig, open_dict

from pinnstorch.utils import pylogger, rich_utils

log = pylogger.get_pylogger(__name__)

def extras(cfg: DictConfig) -> None:
    """Applies optional utilities before the task is started.

    Utilities:
        - Ignoring python warnings
        - Setting tags from command line
        - Rich config printing

    :param cfg: A DictConfig object containing the config tree.
    """
    # return if no `extras` config
    if not cfg.get("extras"):
        log.warning("Extras config not found! <cfg.extras=null>")
        return

    # disable python warnings
    if cfg.extras.get("ignore_warnings"):
        log.info("Disabling python warnings! <cfg.extras.ignore_warnings=True>")
        warnings.filterwarnings("ignore")

    # prompt user to input tags from command line if none are provided in the config
    if cfg.extras.get("enforce_tags"):
        log.info("Enforcing tags! <cfg.extras.enforce_tags=True>")
        rich_utils.enforce_tags(cfg, save_to_file=True)

    # pretty print config tree using Rich library
    if cfg.extras.get("print_config"):
        log.info("Printing config tree with Rich! <cfg.extras.print_config=True>")
        rich_utils.print_config_tree(cfg, resolve=True, save_to_file=True)


def task_wrapper(task_func: Callable) -> Callable:
    """Optional decorator that controls the failure behavior when executing the task function.

    This wrapper can be used to:
        - make sure loggers are closed even if the task function raises an exception (prevents multirun failure)
        - save the exception to a `.log` file
        - mark the run as failed with a dedicated file in the `logs/` folder (so we can find and rerun it later)
        - etc. (adjust depending on your needs)

    Example:
    ```
    @utils.task_wrapper
    def train(cfg: DictConfig) -> Tuple[Dict[str, Any], Dict[str, Any]]:
        ...
        return metric_dict, object_dict
    ```

    :param task_func: The task function to be wrapped.

    :return: The wrapped task function.
    """

    def wrap(
        cfg: DictConfig, read_data_fn: Callable, pde_fn: Callable, output_fn: Callable
    ) -> Tuple[Dict[str, Any], Dict[str, Any]]:
        # execute the task
        try:
            metric_dict, object_dict = task_func(
                cfg=cfg, read_data_fn=read_data_fn, pde_fn=pde_fn, output_fn=output_fn
            )

        # things to do if exception occurs
        except Exception as ex:
            # save exception to `.log` file
            log.exception("")

            # some hyperparameter combinations might be invalid or cause out-of-memory errors
            # so when using hparam search plugins like Optuna, you might want to disable
            # raising the below exception to avoid multirun failure
            raise ex

        # things to always do after either success or exception
        finally:
            # display output dir path in terminal
            log.info(f"Output dir: {cfg.paths.output_dir}")

            # always close wandb run (even if exception occurs so multirun won't fail)
            if find_spec("wandb"):  # check if wandb is installed
                import wandb

                if wandb.run:
                    log.info("Closing wandb!")
                    wandb.finish()

        return metric_dict, object_dict

    return wrap


def get_metric_value(metric_dict: Dict[str, Any], metric_names: list) -> float:
    """Safely retrieves value of the metric logged in LightningModule.

    :param metric_dict: A dict containing metric values.
    :param metric_name: The name of the metric to retrieve.
    :return: The value of the metric.
    """
    for type_metric, list_metrics in metric_names.items():
        if type_metric == "extra_variables":
            prefix = ""
        elif type_metric == "error":
            prefix = "val/error_"

        for metric_name in list_metrics:
            metric_name = f"{prefix}{metric_name}"

            if not metric_name:
                log.info("Metric name is None! Skipping metric value retrieval...")
                continue

            if metric_name not in metric_dict:
                log.info(
                    f"Metric value not found! <metric_name={metric_name}>\n"
                    "Make sure metric name logged in LightningModule is correct!\n"
                    "Make sure `optimized_metric` name in `hparams_search` config is correct!"
                )
            else:
                metric_value = metric_dict[metric_name].item()
                log.info(f"Retrieved metric value! <{metric_name}={metric_value}>")

    return metric_value


def download_file(path, folder_name, filename):
    """Download a file from a given URL and save it to the specified path.

    :param path: Path where the file should be saved.
    :param folder_name: Name of the folder containing the file on the server.
    :param filename: Name of the file to be downloaded.
    """

    url = f"https://storage.googleapis.com/pinns/{folder_name}/{filename}"
    response = requests.get(url, timeout=10)
    if response.status_code == 200:
        os.makedirs(os.path.dirname(path), exist_ok=True)
        with open(path, "wb") as file:
            file.write(response.content)
        log.info("File downloaded successfully.")
    else:
        FileNotFoundError("File download failed.")


def load_data_txt(root_path, file_name):
    """Load text data from a file, downloading it if not already present.

    :param root_path: The root directory where the data file should be located.
    :param file_name: Name of the data file.
    :return: Loaded data as a numpy array.
    """
    path = os.path.join(root_path, file_name)
    if os.path.exists(path):
        log.info("Weights are available.")
    else:
        download_file(path, "irk_weights", file_name)

    return np.float32(np.loadtxt(path, ndmin=2))


def load_data(root_path, file_name):
    """Load data from a MATLAB .mat file, downloading it if not already present.

    :param root_path: The root directory where the data file should be located.
    :param file_name: Name of the data file.
    :return: Loaded data using scipy.io.loadmat function.
    """

    path = os.path.join(root_path, file_name)
    if os.path.exists(path):
        log.info("Data is available.")
    else:
        download_file(path, "data", file_name)

    return scipy.io.loadmat(path)


def set_mode(cfg):
    import torch
    
    if not torch.cuda.is_available():
        log.info("GPU is not found. Using CPU.")
        cfg.trainer.accelerator = "cpu"
        cfg.trainer.devices = 1

    if cfg.model.lazy:
        log.info("Using LazyTensor as backend.")
    
    if cfg.model.cudagraph_compile and cfg.trainer.accelerator != "cpu":
        log.info("Model will be compiled.")
        log.info("Setting optimizer capturable attribute to True.")
        cfg.model.optimizer.capturable = True
        log.info("Disabling automatic optimization.")
        if cfg.trainer.devices is not None and isinstance(cfg.trainer.devices, list):
            if len(cfg.trainer.devices) > 1:
                log.info(
                    f"DDP is not supported for compiled model. Using device {cfg.trainer.devices[0]}"
                )
                cfg.trainer.devices = [cfg.trainer.devices[0]]
        
    elif not cfg.model.cudagraph_compile and cfg.trainer.accelerator != "cpu": 
        log.info("Model will not be compiled.")
        log.info("Setting optimizer capturable attribute to False.")
        cfg.model.optimizer.capturable = False
        log.info("Enabling automatic optimization.")
        if cfg.model.amp:
            with open_dict(cfg):
                cfg.trainer.precision = 16

    elif cfg.trainer.accelerator == "cpu":
        log.info("Model will not be compiled with CUDA Graph.")
        cfg.model.cudagraph_compile = False
        log.info("Setting optimizer capturable attribute to False.")
        cfg.model.optimizer.capturable = False
        log.info("Enabling automatic optimization.")
    
    return cfg


from typing import Dict, List, Union, Optional, Tuple

import torch 
import torch._dynamo

@torch.jit.script
def gradient(dy: torch.Tensor,
             dx: Union[List[torch.Tensor], torch.Tensor],
             ones_like_tensor: Optional[List[Optional[torch.Tensor]]] = None,
             create_graph: bool = True) -> List[torch.Tensor]:
    """Compute the gradient of a tensor `dy` with respect to another tensor `dx`.

    :param dy: The tensor to compute the gradient for.
    :param dx: The tensor with respect to which the gradient is computed.
    :param ones_like_tensor: A tensor with the same shape as `dy`, used for creating the gradient (default is None).
    :param create_graph: Whether to create a computational graph for higher-order gradients (default is True).
    :return: The gradient of `dy` with respect to `dx`.
    """
    if ones_like_tensor is None:
        grad_outputs: List[Optional[torch.Tensor]] = [torch.ones_like(dy)]
    else:
        grad_outputs = ones_like_tensor

    if isinstance(dx, torch.Tensor):
        dx = [dx]

    dy_dx = torch.autograd.grad(
        [dy],
        dx,
        grad_outputs=grad_outputs,
        create_graph=create_graph,
        retain_graph=True,
        allow_unused=False,
    )
    
    grads = [grad if grad is not None else torch.zeros_like(dx[i]) for i, grad in enumerate(dy_dx)]
    return grads

@torch.jit.script
def fwd_gradient(dy: torch.Tensor, 
                 dx: Union[List[torch.Tensor], torch.Tensor],
                 create_graph: bool = True) -> List[torch.Tensor]:
    """Compute the forward gradient of a tensor `dy` with respect to another tensor `dx`.

    :param dy: The tensor to compute the forward gradient for.
    :param dx: The tensor with respect to which the forward gradient is computed.
    :return: The forward gradient of `dy` with respect to `dx`.
    """
        
    if isinstance(dx, torch.Tensor):
        dx = [dx]
    
    ones_like: List[Optional[torch.Tensor]] = [torch.ones_like(dy).requires_grad_(True)]
        
    grads = gradient(dy, dx, ones_like)

    fwd_grads: List[torch.Tensor] = []
    for i, grad in enumerate(grads):
        
        ones_like_ = ones_like[0]
        assert ones_like_ is not None
        
        fwd_grad = gradient(grad, ones_like_, create_graph=create_graph)[0]
        
        if isinstance(fwd_grad, torch.Tensor):
            fwd_grads.append(fwd_grad)
    
    return fwd_grads


from typing import Tuple, List, Dict

import torch

from torch.utils._python_dispatch import _disable_current_modes
from torch._subclasses.fake_tensor import FakeTensor

import functools

def defake(x):
    """
    Convert FakeTensor instances to real, zero-filled tensors with the same meta-data.

    If the input is not a FakeTensor, it is returned unmodified. Otherwise, the function creates a 
    zero tensor with the same size, stride, dtype, device, and requires_grad properties as the FakeTensor. 
    Symbolic sizes and strides are resolved to concrete values if necessary.

    :param x: The input tensor, which can either be a FakeTensor or a regular tensor.
    :return: A zero-filled tensor with the same properties as the input, or the unmodified input if it's not a FakeTensor.
    """
    
    if not isinstance(x, FakeTensor):
        return x
    if x._has_symbolic_sizes_strides:
        size = [
            s.node.shape_env.size_hint(s.node.expr)
            if isinstance(s, torch.SymInt)
            else s
            for s in x.size()
        ]
        stride = [
            s.node.shape_env.size_hint(s.node.expr)
            if isinstance(s, torch.SymInt)
            else s
            for s in x.stride()
        ]
    else:
        size = x.size()
        stride = x.stride()
    y = torch.empty_strided(
        size,
        stride,
        dtype=x.dtype,
        device=x.device,
        requires_grad=x.requires_grad,
    )
    y.requires_grad_(False).zero_()
    return y

def fake_tensor_unsupported(fn):
    """
    A decorator for functions that are incompatible with FakeTensors. 

    This decorator ensures that any FakeTensors in the inputs are converted to real tensors filled with zeros 
    before passing them to the wrapped function. It disables the current dispatch modes temporarily to perform 
    this conversion.

    :param fn: The function to be wrapped.
    :return: The wrapped function that can handle real tensors only.
    """

    @functools.wraps(fn)
    def wrapper(model, inputs, **kwargs):
        with _disable_current_modes():
            inputs = list(map(defake, inputs))
            return fn(model, inputs, **kwargs)

    return wrapper        


def strip_overloads(gm):
    """
    Update the targets of graph nodes in the given Fx graph module to remove overloads.

    Iterates through all nodes in the graph, and if the node target is an OpOverload instance, 
    replaces it with the original operator.

    :param gm: The input Fx graph module containing the nodes to be updated.
    """
    for node in gm.graph.nodes:
        if isinstance(node.target, torch._ops.OpOverload):
            node.target = node.target.overloadpacket
    gm.recompile()

from contextlib import contextmanager

@contextmanager
def _disable_jit_autocast():
    """
    A context manager to temporarily disable JIT autocast mode.

    This is used to turn off autocasting during the JIT compilation process, ensuring that operations are 
    not automatically casted to different data types to improve performance. It saves the current state of the 
    JIT autocast flag, disables autocasting, and then restores the flag to its original state after exiting 
    the context.

    """
    
    old_jit_autocast_flag = torch._C._jit_set_autocast_mode(False)
    try:
        yield
    finally:
        torch._C._jit_set_autocast_mode(old_jit_autocast_flag)


@fake_tensor_unsupported
def jit_compiler(gm: torch.fx.GraphModule, example_inputs: List[torch.Tensor]):
    """
    Compile the given Fx graph module using JIT, handling fake tensors and overloads appropriately.

    This function, wrapped with the @fake_tensor_unsupported decorator, ensures that the graph module is 
    compatible with real tensors only. It disables JIT autocast, strips overloads, and updates graph nodes 
    with certain conditions. Finally, it uses JIT trace to compile the graph module and performs additional 
    optimizations.

    :param gm: The Fx graph module to be compiled.
    :param example_inputs: A list of example inputs to be used during the tracing process.
    :return: The compiled JIT function.
    """
    
    with _disable_jit_autocast():
        strip_overloads(gm)

        for node in gm.graph.nodes:
            if (
                node.target == torch.ops.aten._to_copy
                and len(node.args) == 1
                and len(node.kwargs) == 1
                and "dtype" in node.kwargs
            ):
                node.target = torch.ops.aten.to

        for node in gm.graph.nodes:
            new_kwargs = {}
            for k, v in node.kwargs.items():
                if isinstance(v, torch.device):
                    v = v.type
                new_kwargs[k] = v
            node.kwargs = new_kwargs

        gm.graph.lint()
        gm.recompile()

        fn = torch.jit.trace(gm, example_inputs)
        torch._C._te.remove_unused_self_argument(fn.graph)
        torch._C._jit_pass_remove_mutation(fn.graph)

    return fn


import os
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple

import hydra
import numpy as np
import pytest
import rootutils
import torch
from hydra.core.hydra_config import HydraConfig
from omegaconf import DictConfig, open_dict

import pinnstorch
from pinnstorch.train import train
from tests.helpers.run_if import RunIf


def read_data_fn(root_path):
    """Read and preprocess data from the specified root path.

    :param root_path: The root directory containing the data.
    :return: Processed data will be used in Mesh class.
    """

    data = pinnstorch.utils.load_data(root_path, "KdV.mat")
    exact_u = np.real(data["uu"]).astype("float32")
    return {"u": exact_u}


def pde_fn(outputs, x, extra_variables):
    """Define the partial differential equations (PDEs)."""

    U = outputs["u"]
    U_x = pinnstorch.utils.fwd_gradient(U, x)[0]
    U_xx = pinnstorch.utils.fwd_gradient(U_x, x)[0]
    U_xxx = pinnstorch.utils.fwd_gradient(U_xx, x)[0]
    outputs["f"] = -extra_variables["l1"] * U * U_x - torch.exp(extra_variables["l2"]) * U_xxx
    return outputs


def test_train_fast_dev_run(cfg_train: DictConfig) -> None:
    """Run for 1 train, val and test step.

    :param cfg_train: A DictConfig containing a valid training configuration.
    """
    HydraConfig().set_config(cfg_train)
    with open_dict(cfg_train):
        cfg_train.trainer.fast_dev_run = True
        cfg_train.trainer.accelerator = "cpu"
    train(cfg_train, read_data_fn=read_data_fn, pde_fn=pde_fn, output_fn=None)


@RunIf(min_gpus=1)
def test_train_fast_dev_run_gpu(cfg_train: DictConfig) -> None:
    """Run for 1 train, val and test step on GPU.

    :param cfg_train: A DictConfig containing a valid training configuration.
    """
    HydraConfig().set_config(cfg_train)
    with open_dict(cfg_train):
        cfg_train.trainer.fast_dev_run = True
        cfg_train.trainer.accelerator = "gpu"
    train(cfg_train, read_data_fn=read_data_fn, pde_fn=pde_fn, output_fn=None)


@RunIf(min_gpus=1)
@pytest.mark.slow
def test_train_epoch_gpu_amp(cfg_train: DictConfig) -> None:
    """Train 1 epoch on GPU with mixed-precision.

    :param cfg_train: A DictConfig containing a valid training configuration.
    """
    HydraConfig().set_config(cfg_train)
    with open_dict(cfg_train):
        cfg_train.trainer.max_epochs = 1
        cfg_train.trainer.accelerator = "cpu"
        cfg_train.trainer.precision = 16
    train(cfg_train, read_data_fn=read_data_fn, pde_fn=pde_fn, output_fn=None)


@pytest.mark.slow
def test_train_epoch_double_val_loop(cfg_train: DictConfig) -> None:
    """Train 1 epoch with validation loop twice per epoch.

    :param cfg_train: A DictConfig containing a valid training configuration.
    """
    HydraConfig().set_config(cfg_train)
    with open_dict(cfg_train):
        cfg_train.trainer.max_epochs = 1
        cfg_train.trainer.val_check_interval = 0.5
    train(cfg_train, read_data_fn=read_data_fn, pde_fn=pde_fn, output_fn=None)


@pytest.mark.slow
def test_train_ddp_sim(cfg_train: DictConfig) -> None:
    """Simulate DDP (Distributed Data Parallel) on 2 CPU processes.

    :param cfg_train: A DictConfig containing a valid training configuration.
    """
    HydraConfig().set_config(cfg_train)
    with open_dict(cfg_train):
        cfg_train.trainer.max_epochs = 2
        cfg_train.trainer.accelerator = "cpu"
        cfg_train.trainer.devices = 2
        cfg_train.trainer.strategy = "ddp_spawn"
    train(cfg_train, read_data_fn=read_data_fn, pde_fn=pde_fn, output_fn=None)


@pytest.mark.slow
def test_train_resume(tmp_path: Path, cfg_train: DictConfig) -> None:
    """Run 1 epoch, finish, and resume for another epoch.

    :param tmp_path: The temporary logging path.
    :param cfg_train: A DictConfig containing a valid training configuration.
    """
    with open_dict(cfg_train):
        cfg_train.trainer.max_epochs = 1

    HydraConfig().set_config(cfg_train)
    metric_dict_1, _ = train(cfg_train, read_data_fn=read_data_fn, pde_fn=pde_fn, output_fn=None)

    files = os.listdir(tmp_path / "checkpoints")
    assert "last.ckpt" in files
    assert "epoch_000.ckpt" in files

    with open_dict(cfg_train):
        cfg_train.ckpt_path = str(tmp_path / "checkpoints" / "last.ckpt")
        cfg_train.trainer.max_epochs = 2

    metric_dict_2, _ = train(cfg_train, read_data_fn=read_data_fn, pde_fn=pde_fn, output_fn=None)

    files = os.listdir(tmp_path / "checkpoints")
    assert "epoch_001.ckpt" in files
    assert "epoch_002.ckpt" not in files

    assert metric_dict_1["train/acc"] < metric_dict_2["train/acc"]
    assert metric_dict_1["val/acc"] < metric_dict_2["val/acc"]



from typing import Any, Dict, Optional, Tuple, Callable

import torch
from lightning import LightningDataModule

from .dataloader.dataloader import PINNDataLoader


class PINNDataModule(LightningDataModule):
    """`LightningDataModule` for the PINN dataset.

    This Data Module will hold datasets that should be used for training and Validation.

    A `LightningDataModule` implements 7 key methods:

    ```python
        def prepare_data(self):
        # Things to do on 1 GPU/TPU (not on every GPU/TPU in DDP).
        # Download data, pre-process, split, save to disk, etc...

        def setup(self, stage):
        # Things to do on every process in DDP.
        # Load data, set variables, etc...

        def train_dataloader(self):
        # return train dataloader

        def val_dataloader(self):
        # return validation dataloader

        def test_dataloader(self):
        # return test dataloader

        def predict_dataloader(self):
        # return predict dataloader

        def teardown(self, stage):
        # Called on every process in DDP.
        # Clean up after fit or test.
    ```

    This allows you to share a full dataset without explaining how to download,
    split, transform and process the data.

    Read the docs:
        https://lightning.ai/docs/pytorch/latest/data/datamodule.html
    """
    function_mapping: Dict[str, Callable]
    
    def __init__(
        self,
        train_datasets,
        val_dataset,
        test_dataset=None,
        pred_dataset=None,
        batch_size: int = None,
        num_workers: int = 0,
        pin_memory: bool = False,
    ) -> None:
        """Initialize a `PINNDataModule`.

        :param train_datasets: train datasets.
        :param val_dataset: validation dataset.
        """
        super().__init__()

        self.train_datasets = train_datasets
        self.val_dataset = val_dataset
        self.test_dataset = test_dataset
        self.pred_dataset = pred_dataset

        if isinstance(val_dataset, list):
            raise "Validation dataset cannot be a list."

        self.batch_size = batch_size

        # this line allows to access init params with 'self.hparams' attribute
        # also ensures init params will be stored in ckpt
        #self.save_hyperparameters(logger=False)

        self.data_train = None
        self.data_val = self.val_dataset
        self.data_test = (
            self.test_dataset if self.test_dataset else self.val_dataset
        )
        self.data_pred = (
            self.pred_dataset if self.pred_dataset else self.val_dataset
        )

        self.function_mapping = torch.jit.annotate(Dict[str, Callable], {})

    def setup(self, stage: Optional[str] = None) -> None:
        """Load data. Set variables: `self.data_train`, `self.data_val`, `self.data_test`.

        This method is called by Lightning before `trainer.fit()`, `trainer.validate()`, `trainer.test()`, and
        `trainer.predict()`, so be careful not to execute things like random split twice! Also, it is called after
        `self.prepare_data()` and there is a barrier in between which ensures that all the processes proceed to
        `self.setup()` once the data is prepared and available for use.

        :param stage: The stage to setup. Either `"fit"`, `"validate"`, `"test"`, or `"predict"`. Defaults to ``None``.
        """

        # load and split datasets only if not loaded already
        if not self.data_train:
            self.data_train = {}
            self.set_mode_for_discrete_mesh()

            for train_dataset in self.train_datasets:
                self.data_train[str(train_dataset.loss_fn)] = PINNDataLoader(
                    train_dataset, batch_size=self.batch_size, ignore=True, shuffle=True
                )
                self.function_mapping[str(train_dataset.loss_fn)] = train_dataset.loss_fn

    def set_mode_for_discrete_mesh(self):
        """This function will figuere out which training datasets are for discrete.

        Then set the mode value that will be used for Runge–Kutta methods
        """

        mesh_idx = [
            (train_dataset.idx_t, train_dataset)
            for train_dataset in self.train_datasets
            if type(train_dataset).__name__ == "DiscreteMeshSampler"
        ]

        mesh_idx = sorted(mesh_idx, key=lambda x: x[0])

        if len(mesh_idx) == 1:
            mesh_idx[0][1].mode = "forward_discrete"
        elif len(mesh_idx) == 2:
            mesh_idx[0][1].mode = "inverse_discrete_1"
            mesh_idx[1][1].mode = "inverse_discrete_2"

        mesh_idx.clear()

    def train_dataloader(self):
        """Create and return the train dataloader.

        :return: The train dataloader.
        """
        return self.data_train

    def val_dataloader(self):
        """Create and return the validation dataloader. `self.loss_fn` and `self.solution` will be
        used in `PINNModule`. Lightning does not allow multiple val datasets.

        :return: The validation dataloader.
        """

        self.loss_fn = self.data_val.loss_fn
        self.function_mapping[str(self.loss_fn)] = self.loss_fn
        self.solution_names = self.data_val.solution_names

        return PINNDataLoader(self.data_val, batch_size=self.batch_size, shuffle=False)

    def test_dataloader(self):
        """Create and return the test dataloader. `self.loss_fn` and `self.solution` will be used
        in `PINNModule`. Lightning does not allow multiple val datasets.

        :return: The test dataloader.
        """
        self.loss_fn = self.test_dataset.loss_fn if self.test_dataset else self.val_dataset.loss_fn
        self.function_mapping[str(self.loss_fn)] = self.loss_fn
        self.solution_names = (
            self.test_dataset.solution_names
            if self.test_dataset
            else self.val_dataset.solution_names
        )

        return PINNDataLoader(self.data_test, batch_size=self.batch_size, shuffle=False)

    def predict_dataloader(self):
        """Create and return the test dataloader. `self.loss_fn` and `self.solution` will be used
        in `PINNModule`. Lightning does not allow multiple val datasets.

        :return: The test dataloader.
        """
        self.loss_fn = self.pred_dataset.loss_fn if self.pred_dataset else self.val_dataset.loss_fn
        self.function_mapping[str(self.loss_fn)] = self.loss_fn
        self.solution_names = (
            self.pred_dataset.solution_names
            if self.pred_dataset
            else self.val_dataset.solution_names
        )

        return PINNDataLoader(self.data_pred, batch_size=self.batch_size, shuffle=False)

    def teardown(self, stage: Optional[str] = None) -> None:
        """Lightning hook for cleaning up after `trainer.fit()`, `trainer.validate()`,
        `trainer.test()`, and `trainer.predict()`.

        :param stage: The stage being torn down. Either `"fit"`, `"validate"`, `"test"`, or `"predict"`.
            Defaults to ``None``.
        """
        pass

    def state_dict(self) -> Dict[Any, Any]:
        """Called when saving a checkpoint. Implement to generate and save the datamodule state.

        :return: A dictionary containing the datamodule state that you want to save.
        """
        return {}

    def load_state_dict(self, state_dict: Dict[str, Any]) -> None:
        """Called when loading a checkpoint. Implement to reload datamodule state given datamodule
        `state_dict()`.

        :param state_dict: The datamodule state returned by `self.state_dict()`.
        """
        pass
        


if __name__ == "__main__":
    _ = PINNDataModule(None, None)



from typing import Dict, List, Union, Optional, Tuple

import torch

def sse(loss: torch.Tensor,
        preds: Dict[str, torch.Tensor],
        target: Union[Dict[str, torch.Tensor], None] = None,
        keys: Union[List[str], None] = None,
        mid: Union[int, None] = None) -> torch.Tensor:
    """Calculate the sum of squared errors (SSE) loss for given predictions and optional targets.

    :param loss: Loss variable.
    :param preds: Dictionary containing prediction tensors for different keys.
    :param target: Dictionary containing target tensors (optional).
    :param keys: List of keys for which to calculate SSE loss (optional).
    :param mid: Index to separate predictions for mid-point calculation (optional).
    :return: Calculated SSE loss.
    """
    
    if keys is None:
        return loss

    for key in keys:
        if target is None and mid is None:
            loss = loss + torch.sum(torch.square(preds[key]))
        elif target is None and mid is not None:
            loss = loss + torch.sum(torch.square(preds[key][:mid] - preds[key][mid:]))
        elif target is not None:
            loss = loss + torch.sum(torch.square(preds[key] - target[key]))

    return loss

def mse(loss: torch.Tensor,
        preds: Dict[str, torch.Tensor],
        target: Union[Dict[str, torch.Tensor], None] = None,
        keys: Union[List[str], None] = None,
        mid: Union[int, None] = None) -> torch.Tensor:
    """Calculate the mean squared error (MSE) loss for given predictions and optional targets.

    :param loss: Loss variable.
    :param preds: Dictionary containing prediction tensors for different keys.
    :param target: Dictionary containing target tensors (optional).
    :param keys: List of keys for which to calculate SSE loss (optional).
    :param mid: Index to separate predictions for mid-point calculation (optional).
    :return: Calculated MSE loss.
    """

    if keys is None:
        return loss

    for key in keys:
        if target is None and mid is None:
            loss = loss + torch.mean(torch.square(preds[key]))
        elif target is None and mid is not None:
            loss = loss + torch.mean(torch.square(preds[key][:mid] - preds[key][mid:]))
        elif target is not None:
            loss = loss + torch.mean(torch.square(preds[key] - target[key]))

    return loss


def relative_l2_error(preds, target):
    """Calculate the relative L2 error between predictions and target tensors.

    :param preds: Predicted tensors.
    :param target: Target tensors.
    :return: Relative L2 error value.
    """
    l2_norm_error = torch.mean((preds - target)**2)
    l2_norm_target = torch.mean((target)**2)
    #print(l2_norm_error, l2_norm_target, relative_l2)
    if l2_norm_target == 0:
        return l2_norm_error
    return torch.sqrt(l2_norm_error/l2_norm_target)


def fix_extra_variables(extra_variables):
    """Convert extra variables to torch tensors with gradient tracking. These variables are
    trainables in inverse problems.

    :param extra_variables: Dictionary of extra variables to be converted.
    :return: Dictionary of converted extra variables as torch tensors with gradients.
    """
    if extra_variables is None:
        return None
    extra_variables_parameters = {}
    for key in extra_variables:
        extra_variables_parameters[key] = torch.tensor(
            [extra_variables[key]], dtype=torch.float32, requires_grad=True
        )
    extra_variables_parameters = torch.nn.ParameterDict(extra_variables_parameters)
    return extra_variables_parameters


def requires_grad(batch, enable_grad=True):
    """Set the requires_grad attribute for tensors in the input list.

    :param batch: A batch containing spatial, time, solutions.
    :param enable_grad: Boolean indicating whether to enable requires_grad or not.
    :return: Modified list of tensors and tensor.
    """
    spatial, time, solutions = batch
    if time is not None:
        time = time.requires_grad_(enable_grad)
    spatial = [x_.requires_grad_(enable_grad) for x_ in spatial]
    solutions = {
        solution_name: solutions[solution_name].requires_grad_(enable_grad)
        for solution_name in solutions.keys()
    }

    return spatial, time, solutions


def set_requires_grad(x: List[torch.Tensor],
                      t: torch.Tensor,
                      enable_grad: bool=True) -> Tuple[List[torch.Tensor], torch.Tensor]:
    """Set the requires_grad attribute for tensors in the input list.

    :param x: List of tensors to modify requires_grad attribute.
    :param t: Tensor to modify requires_grad attribute.
    :param enable_grad: Boolean indicating whether to enable requires_grad or not.
    :return: Modified list of tensors and tensor.
    """
    if t is not None:
        t = t.requires_grad_(enable_grad)
    x = [x_.requires_grad_(enable_grad) for x_ in x]

    return x, t

def fix_predictions(preds_list: List[Dict[str, torch.Tensor]]) -> Dict[str, torch.Tensor]:
    """It will concatenate predictions for each solution 
    into a tensor and return a dictionary including them.

    :param preds_list: List of dictionary of predictions.
    :return: A dictionary containing the predictions.
    """
    preds_dict = {}
    for preds in preds_list:
        for sol_key, pred in preds.items():
            if sol_key in preds_dict.keys():
                preds_dict[sol_key] = torch.cat((preds_dict[sol_key], pred.detach()), 0)
            else:
                preds_dict[sol_key] = pred.detach()
    return preds_dict





import numpy as np
import torch

from .sampler_base import SamplerBase
from pinnstorch.utils import set_requires_grad


class InitialCondition(SamplerBase):
    """Initialize initial boundary condition."""

    def __init__(self, mesh, num_sample=None, solution=None, initial_fun=None):
        """Initialize an InitialCondition object for sampling initial condition data.

        :param mesh: Mesh object containing spatial and time domain information.
        :param num_sample: Number of samples.
        :param solution: List of solution variable names.
        :param initial_fun: Function to generate initial conditions (optional).
        """
        super().__init__()

        self.solution_names = solution

        (self.spatial_domain, self.time_domain, self.solution) = mesh.on_initial_boundary(
            self.solution_names
        )

        if initial_fun:
            self.solution = initial_fun(self.spatial_domain)

        (
            self.spatial_domain_sampled,
            self.time_domain_sampled,
            self.solution_sampled,
        ) = self.sample_mesh(num_sample, (self.spatial_domain, self.time_domain, self.solution))

        self.spatial_domain_sampled = list(torch.split(self.spatial_domain_sampled, (1), dim=1))
        self.solution_sampled = list(torch.split(self.solution_sampled, (1), dim=1))

    def sample_mesh(self, num_sample, flatten_mesh):
        """Sample the mesh data for training. If num_sample is not defined the whole points will be
        selected.

        :param num_sample: Number of samples to generate.
        :param flatten_mesh: Flattened mesh data.
        :return: Sampled spatial, time, and solution data.
        """
        flatten_mesh = list(flatten_mesh)
        concatenated_solutions = [
            flatten_mesh[2][solution_name] for solution_name in self.solution_names
        ]
        flatten_mesh[2] = np.concatenate(concatenated_solutions, axis=-1)

        if num_sample is None:
            return self.convert_to_tensor(flatten_mesh)
        else:
            idx = np.random.choice(range(flatten_mesh[0].shape[0]), num_sample, replace=False)
            return self.convert_to_tensor(
                (flatten_mesh[0][idx, :], flatten_mesh[1][idx, :], flatten_mesh[2][idx, :])
            )

    def _loss_fn(self, inputs, loss):
        """Compute the loss function based on inputs and functions.

        :param inputs: Input data for computing the loss.
        :param loss: Loss variable.
        :param functions: Additional functions required for loss computation.
        :return: Loss variable and outputs dict from the forward pass.
        """

        x, t, u = inputs

        outputs = self.functions["forward"](x, t)
        loss = self.functions["loss_fn"](loss, outputs, u, keys=self.solution_names)

        return loss, outputs



from typing import Any, Callable, Dict, List, Tuple, Union

import os
import time
import torch
import torch._lazy
import torch._lazy.ts_backend
import torch.distributed as dist
import torch.cuda.amp as amp
import numpy as np

from lightning import LightningModule
from lightning.pytorch.utilities import move_data_to_device
from torchmetrics import MeanMetric, MinMetric
from torchmetrics.classification.accuracy import Accuracy
from lightning.pytorch.plugins.precision.amp import MixedPrecisionPlugin

from pinnstorch.utils.gradient_fn import fwd_gradient, gradient
from pinnstorch.utils.module_fn import (
    fix_extra_variables,
    mse,
    relative_l2_error,
    set_requires_grad,
    sse,
)

from pinnstorch.utils import get_pylogger

log = get_pylogger(__name__)

class PINNModule(LightningModule):
    """
    A LightningModule for training Physics-Informed Neural Networks (PINNs) to solve partial 
    differential equations (PDEs).
    """

    function_mapping: Dict[str, Callable]
    net: torch.nn.Module
    
    def __init__(
        self,
        net: torch.nn.Module,
        pde_fn,
        optimizer: torch.optim.Optimizer = torch.optim.Adam,
        scheduler: torch.optim.lr_scheduler = None,
        scaler: torch.cuda.amp.GradScaler = None,
        loss_fn: str = "sse",
        extra_variables: Dict[str, Any] = None,
        output_fn = None,
        runge_kutta=None,
        cudagraph_compile: bool = False,
        jit_compile: bool = True,
        amp: bool = True,
        lazy: bool = False,
        inline: bool = False
    ) -> None:
        """
        Initialize the PINNModule.

        Sets up the model, loss function, optional features like AMP, JIT compilation, and CUDA Graphs,
        and other utilities for training and evaluating a Physics-Informed Neural Network.

        :param net: The neural network model.
        :param pde_fn: The function representing the PDE to solve.
        :param optimizer: The optimizer for training.
        :param scheduler: Optional learning rate scheduler.
        :param scaler: Optional gradient scaler for AMP.
        :param loss_fn: The loss function to use, either "sse" or "mse".
        :param extra_variables: Optional extra variables in inverse problems.
        :param output_fn: Optional function to process the model's output.
        :param runge_kutta: Optional Runge-Kutta method for solving PDEs.
        :param cudagraph_compile: Flag to enable CUDA Graph compilation. It works only with a single GPU.
        :param jit_compile: Flag to enable JIT compilation.
        :param amp: Flag to enable Automatic Mixed Precision (AMP).
        :param lazy: Flag to enable the use of LazyTensors.
        :param inline: Flag to enable inline mode in JIT compilation.
        """
        super().__init__()

        # this line allows to access init params with 'self.hparams' attribute
        # also ensures init params will be stored in ckpt
        self.save_hyperparameters(logger=False, ignore=["net", "runge_kutta", "scaler"])

        self.net = net
        self.capture_end = False
        self.extra_variables: torch.nn.ParameterDict = fix_extra_variables(extra_variables)
        self.pde_fn = pde_fn
        self.output_fn = output_fn
        self.rk = runge_kutta
        self.cudagraph_compile = cudagraph_compile
        self.jit_compile = jit_compile
        self.preds_dict = {}
        self.lazy = lazy
        self.amp = amp
        self.opt = None
        self.scaler = scaler
        self.times = []
        self.times_batch = []
        self.xla = False

        if loss_fn == "sse":
            self.loss_fn = sse
        elif loss_fn == "mse":
            self.loss_fn = mse

        if self.amp:
            torch._C._jit_set_autocast_mode(True)

        if self.cudagraph_compile:
            self.automatic_optimization = False
            
        if self.jit_compile and self.cudagraph_compile:
            if inline:
                torch._C._debug_set_autodiff_subgraph_inlining(False)
                torch._C._jit_set_nvfuser_single_node_mode(True)
            else:
                torch._C._jit_set_nvfuser_enabled(True)
        
        if self.lazy:
            torch._lazy.ts_backend.init()

        
        # for averaging loss across batches
        self.train_loss = MeanMetric()
        self.val_loss = MeanMetric()
        self.val_error = MeanMetric()
        self.test_loss = MeanMetric()
        self.test_error = MeanMetric()

        # for tracking best so far validation accuracy
        self.val_loss_best = MinMetric()

        # for using in each loss function of each condition.
        self.functions = {
            "runge_kutta": self.rk,
            "forward": self.forward,
            "pde_fn": self.pde_fn,
            "output_fn": self.output_fn,
            "extra_variables": self.extra_variables,
            "loss_fn": self.loss_fn,
            "net": self.net,
            "jit_compile": self.jit_compile,
        }

    def forward(self, spatial: List[torch.Tensor], time: torch.Tensor) -> Dict[str, torch.Tensor]:
        """Perform a forward pass through the model `self.net`.

        :param spatial: List of input spatial tensors.
        :param time: Input tensor representing time.
        :return: A tensor of solutions.
        """
        outputs = self.net(spatial, time)
        if self.output_fn is not None:
            outputs = self.output_fn(outputs, *spatial, time)
        return outputs

    def on_fit_start(self) -> None:
        """
        Initialize the function mapping at the start of the fitting process.

        :return: None
        """
        
        self.function_mapping = self.trainer.datamodule.function_mapping
        
    def on_train_start(self) -> None:
        """
        Lightning hook that is called when training begins.

        By default lightning executes validation step sanity checks before training starts,
        so it's worth to make sure validation metrics don't store results from these checks
        
        :return: None
        """
        
        self.val_loss.reset()
        self.val_error.reset()
        self.val_loss_best.reset()

        self.functions["batch_size"] = True if self.trainer.datamodule.batch_size else False
        
        if self.lazy: 
            device = 'lazy'
        elif self.xla:
            device = 'xla'
        else:
            device = self.device
        self.net = self.net.to(device)
        self.train_loss = self.train_loss.to(device)
        if self.extra_variables:
            self.extra_variables = self.extra_variables.to(device)
        if self.rk:
            self.rk = self.rk.to(device)

    def on_train_batch_start(self):
        """
        Prepare the training loss tensor for the upcoming training batch.
        This method is called at the beginning of each training batch
        to ensure the `train_loss` tensor is on the appropriate device.
        
        :return: None
        """
        
        if self.lazy: 
            device = 'lazy'
        elif self.xla:
            device = 'xla'
        else:
            device = self.device
            
        self.train_loss = self.train_loss.to(device)
            
    def transfer_batch_to_device(self, batch, device, dataloader_idx):
        """
        Perform moving the data to the target device.
        If we do not use CUDA graphs we need to move the batch to self.device for
        every batch, but for capturing CUDA graphs, we have to do it one time.

        :param batch: A batch of data (a tuple) containing the
        :param device: The target device
        :param dataloader_idx: The index of dataloader 

        :return: The batch in target device.
        """
        
        if self.lazy:
            device = 'lazy'

        if self.xla:
            device = 'xla'

        if self.automatic_optimization:
            # If we do not use CUDA graphs we need to move the batch to self.device.
            batch = super().transfer_batch_to_device(batch, device, dataloader_idx)
        else:
            # If we use CUDA graphs we need to move the batch to self.device just for the first time.
            if self.capture_end is False:
                batch = super().transfer_batch_to_device(batch, device, dataloader_idx)
            else:
                if self.trainer.datamodule.batch_size and self.val_stage is False:
                    self.copy_batch(batch)
                elif self.trainer.datamodule.batch_size and self.val_stage is True:
                    batch = super().transfer_batch_to_device(batch, device, dataloader_idx)
            
        return batch

   
    def on_before_backward(self, loss):
        """    
        This method is invoked before the backward pass in the training process.
        If both AMP and CUDA Graph compilation are enabled,
        it scales the loss value using the model's scaler to facilitate a more numerically stable 
        and efficient backward pass. Otherwise, the loss remains unmodified.
    
        :param loss: The original loss value calculated from the forward pass.
        :return: The potentially scaled loss value, depending on the AMP and CUDA Graph compilation settings.
        """
        
        if self.amp and self.cudagraph_compile:
            return self.scaler.scale(loss)
        else:
            return loss
    
    def capture_graph(self, batch):
        """
        Capture the computation graph after warming up with 11 iterations.
    
        This method initiates a warm-up phase of 11 training iterations and then captures the CUDA computation graph for 
        the subsequent iterations. This is particularly useful for optimizing GPU computations by reusing the 
        precompiled computation graph.
    
        The warm-up and graph capturing are performed on a separate CUDA stream to avoid interfering with the ongoing 
        computations on the default stream. The captured graph encompasses the forward and backward passes, as well 
        as the optimizer step, making the entire training iteration highly efficient.
    
        :param batch: The input batch of data to be used for warm-up and graph capturing.
        :return: None
        """
        
        self.capture_time = time.time()
        self.opt = self.optimizers()
        
        self.static_batch = batch
        self.batch_idx = 0
        
        self.s = torch.cuda.Stream()
        self.s.wait_stream(torch.cuda.current_stream())

        with torch.cuda.stream(self.s):
            for i in range(11):
                self.opt.zero_grad(set_to_none=True)
                with torch.autocast(device_type='cuda', enabled = self.amp):
                    loss, pred = self.model_step(self.static_batch)
                    
                self.manual_backward(loss)
                if self.amp:
                    self.scaler.step(self.opt)
                    self.scaler.update()
                else:
                    self.opt.step()

        torch.cuda.current_stream().wait_stream(self.s)
        
        self.g = torch.cuda.CUDAGraph()
        self.opt.zero_grad(set_to_none=True)
        with torch.cuda.graph(self.g):
            with torch.autocast(device_type='cuda', enabled = self.amp):
                self.static_loss, self.static_pred = self.model_step(self.static_batch)
            self.manual_backward(self.static_loss)
            if self.amp is False:
                self.opt.step()

        self.batch_idx = 1
        self.capture_end = True
        print('Capture Time', time.time() - self.capture_time)
    
    def copy_batch(self, batch) -> None:
        """
        Fills the graph's input memory with new data to compute on. If the batch_size is not
        specified, the model uses all the available data, and there is no need to copy the data.

        :param batch: A batch of data (a tuple) containing the input tensor of images and target labels.
        """
        
        import time as tm
        st = tm.time()
        
        for key in batch:
            spatial, time, solution = self.static_batch[key]
            spatial_new, time_new, solution_new = batch[key]
            time = time.requires_grad_(False).copy_(time_new)
            x = [
                spatial_.requires_grad_(False).copy_(spatial_new[i])
                for i, spatial_ in enumerate(spatial)
            ]
            if solution_new is not None:
                solution = {
                    key_sol: solution[key_sol].copy_(solution_new[key_sol])
                    for key_sol in solution_new
                }
        
        self.times_batch.append(tm.time() - st)
        
    def model_step(
        self,
        batch: Dict[
            str,
            Tuple[torch.Tensor, torch.Tensor, torch.Tensor]
        ],
    ) -> Tuple[torch.Tensor, Dict[str, torch.Tensor]]:
        """Perform a single model step on a batch of data.

        :param batch: A batch of data (a tuple) containing the
        input tensor of different conditions and data.

        :return: A tuple containing (in order):
            - A tensor of losses.
            - A dictionary of predictions.
        """
        
        loss = 0.0
        
        for loss_fn, data in batch.items():
            x, t, u = data
            x, t = set_requires_grad(x, t, True)
            loss, preds = self.function_mapping[loss_fn](data, loss, self.functions)
        return loss, preds
        

    def training_step(
        self,
        batch: Dict[
            str,
            Tuple[torch.Tensor, torch.Tensor, torch.Tensor]
        ],
        batch_idx: int,
    ) -> torch.Tensor:
        """Perform a single training step on a batch of data from the training set. self.g.replay()
        includes forward, backward, and step.

        :param batch: A batch of data (a tuple) containing the input tensor of images and target
            labels.
        :param batch_idx: The index of the current batch.
        :return: A tensor of losses between model predictions and targets.
        """
        self.start_time = time.time()
        
        if self.cudagraph_compile:
            if batch_idx == 0 and not self.capture_end:
                self.capture_graph(batch)
            else:
                self.g.replay()
                if self.amp:
                    self.scaler.step(self.opt)
                    self.scaler.update()

            # update and log metrics
            self.train_loss(self.static_loss)
            self.log("train/loss", self.train_loss, on_step=False, on_epoch=True, prog_bar=True, sync_dist=False)

        else:
                
            loss, pred = self.model_step(batch)            
            
            # update and log metrics    
            self.train_loss(loss)
            self.log("train/loss", self.train_loss, on_step=True, on_epoch=True, prog_bar=True, sync_dist=False)
            
            return loss


    def on_train_batch_start(self, batch, batch_idx):
        """
        Set validation flag to False and record start time at a specific training step.
    
        :param batch: The current batch of data being processed.
        :param batch_idx: The index of the current batch.
        :return: None
        """
        self.functions['val'] = False

    def on_train_batch_end(self, batch_output, batch, batch_idx):
        """
        Log the time taken for a specific training step and append the time to the times list.

        :param batch_output: The output of the training batch.
        :param batch: The current batch of data being processed.
        :param batch_idx: The index of the current batch.
        :return: None
        """
        self.times.append(time.time() - self.start_time)
        if self.lazy:
            torch._lazy.mark_step()
    
    def on_train_epoch_end(self):
        """Called at the end of each training epoch.

        If extra_variables are provided, logs each extra variable's value to the progress bar.
        """
        if self.extra_variables:
            for key in self.extra_variables:
                self.log(key, self.extra_variables[key], prog_bar=True, sync_dist=False)
        
        
    def on_validation_start(self):
        """
        Set the validation stage flag to True at the start of the validation phase.

        :return: None
        """
        
        self.val_stage = True

    def on_validation_end(self):
        """
        Reset the validation stage flag to False at the end of the validation phase.

        :return: None
        """
        
        self.val_stage = False

    def eval_step(
        self, batch: Tuple[torch.Tensor, torch.Tensor, torch.Tensor]
    ) -> Tuple[torch.Tensor, Dict[str, torch.Tensor], Dict[str, torch.Tensor]]:
        """Perform a single evaluation step on a batch of data.

        :param batch: A batch of data containing input tensors and conditions.
        :return: A tuple containing loss, error dictionary, and predictions.
        """

        # `loss_fn` and `solution` are passed from PINNDataModule class.
        loss_fn = self.trainer.datamodule.loss_fn
        solution = self.trainer.datamodule.solution_names
        
        
        self.functions['val'] = True
        
        # In evalutation, because we do not use captured graph,
        # we need to transfer batch manually.
        batch = super().transfer_batch_to_device(batch, self.device, 0)
        _, _, u = batch

        with torch.set_grad_enabled(True):
            loss, preds = self.model_step({str(loss_fn): batch})

        if self.rk:
            error_dict = {
                solution_name: relative_l2_error(
                    preds[solution_name][:, -1][:, None], u[solution_name]
                )
                for solution_name in solution
            }
            for solution_name in solution:
                print(preds[solution_name][:, -1][:, None].shape, u[solution_name].shape)

        else:
            error_dict = {
                solution_name: relative_l2_error(preds[solution_name], u[solution_name])
                for solution_name in solution
            }
        
        return loss, error_dict, preds

    def validation_step(
        self,
        batch: Tuple[torch.Tensor, torch.Tensor, torch.Tensor],
        batch_idx: int,
    ) -> None:
        """
        Perform a single validation step on a batch of data from the validation set.

        :param batch: A batch of data.
        :param batch_idx: The index of the current batch.
        """
        loss, error_dict, preds = self.eval_step(batch)
        
        self.val_loss(loss)
        self.log("val/loss", self.val_loss, on_step=False, on_epoch=True, prog_bar=True, sync_dist=False)

        for solution_name, error in error_dict.items():
            self.val_error(error)
            self.log(f"val/error_{solution_name}", error, prog_bar=True, sync_dist=False)

    
    def on_validation_epoch_end(self) -> None:
        """
        Lightning hook that is called when a validation epoch ends.

        :return: None
        """
        
        loss = self.val_loss.compute()
        self.val_loss_best(loss)
        self.log("val/loss_best", self.val_loss_best.compute(), sync_dist=False, prog_bar=True)

    def test_step(
        self,
        batch: Tuple[torch.Tensor, torch.Tensor, torch.Tensor],
        batch_idx: int,
    ) -> None:
        """
        Perform a single test step on a batch of data from the test set.

        :param batch: A batch of data.
        :param batch_idx: The index of the current batch.
        """

        loss, error_dict, preds = self.eval_step(batch)

        # update and log metrics
        self.test_loss(loss)
        self.log("test/loss", loss, on_step=False, on_epoch=True, prog_bar=True, sync_dist=False)

        for solution_name, error in error_dict.items():
            self.test_error(error)
            self.log(f"test/error_{solution_name}", error, prog_bar=True, sync_dist=False)

    def predict_step(
        self,
        batch: Tuple[torch.Tensor, torch.Tensor, torch.Tensor],
        batch_idx: int,
    ) -> None:
        """
        Perform a single predict step on a batch of data from the prediction set.

        :param batch: A batch of data.
        :param batch_idx: The index of the current batch.
        """

        loss, error, preds = self.eval_step(batch)

        return preds

    def configure_optimizers(self) -> Dict[str, Any]:
        """
        Configures optimizers and learning-rate schedulers to be used for training.

        :return: A dict containing the configured optimizers and learning-rate schedulers to be used for training.
        """
        optimizer = self.hparams.optimizer(params=self.parameters())
        if self.hparams.scheduler is not None:
            self.scheduler = self.hparams.scheduler(optimizer=optimizer)
            return {
                "optimizer": optimizer,
                "lr_scheduler": {
                    "scheduler": self.scheduler,
                    "monitor": "val/loss",
                    "interval": "epoch",
                    "frequency": 1,
                },
            }
        return {"optimizer": optimizer}


if __name__ == "__main__":
    _ = PINNModule(None, None, None)



import logging

from typing import Dict, List, Union, Optional

import numpy as np
import torch

from pinnstorch.utils import load_data_txt

log = logging.getLogger(__name__)


class RungeKutta(torch.nn.Module):
    """Initialize Runge Kutta method based on time interval and q."""

    def __init__(self, root_dir, t1: int, t2: int, time_domain, q: int = None):
        """Initialize a RungeKutta object for solving differential equations using Implicit Runge-
        Kutta methods.

        :param root_dir: Root directory where the weights data is stored.
        :param t1: Start index of the time domain.
        :param t2: End index of the time domain.
        :param time_domain: TimeDomain class representing the time domain.
        :param q: Order of the Implicit Runge-Kutta method. If not provided, it is automatically
            calculated.
        """
        super().__init__()

        dt = np.array(time_domain[t2] - time_domain[t1]).astype(np.float32)

        self.register_buffer("dt", torch.from_numpy(dt))

        if q is None:
            q = int(np.ceil(0.5 * np.log(np.finfo(float).eps) / np.log(dt)))

        self.load_irk_weights(root_dir, q)

    def load_irk_weights(self, root_dir, q: int) -> None:
        """Load the weights and coefficients for the Implicit Runge-Kutta method and save in the
        dictionary.

        :param root_dir: Root directory where the weights data is stored.
        :param q: Order of the Implicit Runge-Kutta method.
        """
        file_name = "Butcher_IRK%d.txt" % q
        tmp = load_data_txt(root_dir, file_name)

        weights = np.reshape(tmp[0 : q**2 + q], (q + 1, q)).astype(np.float32)
        
        self.register_buffer("alpha", torch.from_numpy(weights[0:-1, :].T))
        self.register_buffer("beta", torch.from_numpy(weights[-1:, :].T))
        self.register_buffer("weights", torch.from_numpy(weights.T))
        
        self.IRK_times = tmp[q**2 + q :]

    def forward(self,
                outputs: Dict[str, torch.Tensor],
                mode: str,
                solution_names: List[str],
                collection_points_names: List[str]):
        """Perform a forward step using the Runge-Kutta method for solving differential equations.

        :param outputs: Dictionary containing solution tensors and other variables.
        :param mode: The mode of the forward step, e.g., "inverse_discrete_1",
            "inverse_discrete_2", "forward_discrete".
        :param solution_names: List of keys for solution variables.
        :param collection_points_names: List of keys for collection point variables.
        :return: Dictionary with updated solution tensors after the forward step.
        """

        for solution_name, collection_points_name in zip(solution_names, collection_points_names):
            if mode == "inverse_discrete_1":
                outputs[solution_name] = outputs[solution_name] - self.dt * torch.matmul(
                    outputs[collection_points_name], self.alpha
                )

            elif mode == "inverse_discrete_2":
                outputs[solution_name] = outputs[solution_name] + self.dt * torch.matmul(
                    outputs[collection_points_name], (self.beta - self.alpha)
                )

            elif mode == "forward_discrete":
                outputs[solution_name] = outputs[solution_name] - self.dt * torch.matmul(
                    outputs[collection_points_name], self.weights
                )

        return outputs


if __name__ == "__main__":
    _ = RungeKutta(None, None, None, None)



import numpy as np


class Interval:
    """Initialize a 1D spatial interval."""

    def __init__(self, x_interval, shape):
        """Initialize an Interval object to represent a 1D spatial interval.

        :param x_interval: A tuple or list representing the spatial interval [start_x, end_x].
        :param shape: The number of points in the spatial interval.
        """

        self.x_interval = x_interval
        self.shape = shape

    def generate_mesh(self, t_points: int):
        """Generate a mesh for the 1D spatial interval.

        :param t_points: The number of time points in the mesh.
        :return: A mesh for the 1D spatial interval.
        """

        x = np.linspace(self.x_interval[0], self.x_interval[1], num=self.shape[0])
        self.mesh = np.tile(x, (t_points, 1)).T[:, :, None]

        return self.mesh

    def __len__(self):
        """Get the length of the interval mesh.

        :return: The number of points in the interval mesh.
        """

        return len(self.mesh)

    def __getitem__(self, idx):
        """Get a specific point from the interval mesh using indexing.

        :param idx: The index of the desired point.
        :return: The point value at the specified index.
        """

        return self.mesh[idx, 0]


class Rectangle:
    """Initialize a 2D spatial domain."""

    def __init__(self, x_interval, y_interval, shape):
        """Initialize a Rectangle object to represent a 2D spatial rectangle.

        :param x_interval: A tuple representing the x-axis interval [start_x, end_x].
        :param y_interval: A tuple representing the y-axis interval [start_y, end_y].
        :param shape: The number of points along each axis in the rectangle.
        """

        self.x_interval = x_interval
        self.y_interval = y_interval
        self.shape = shape

    def generate_mesh(self, t_points):
        """Generate a mesh for the 2D spatial rectangle.

        :param t_points: The number of time points in the mesh.
        :return: A mesh for the 2D spatial rectangle.
        """

        x = np.linspace(self.x_interval[0], self.x_interval[1], num=self.shape[0])
        y = np.linspace(self.y_interval[0], self.y_interval[1], num=self.shape[1])

        self.xx, self.yy = np.meshgrid(x, y)
        self.spatial_mesh = np.stack((self.xx.flatten(), self.yy.flatten()), 1)

        self.mesh = np.zeros((np.prod(self.shape), t_points, 2))

        for i in range(2):
            self.mesh[:, :, i] = np.tile(self.spatial_mesh[:, i], (t_points, 1)).T

        return self.mesh

    def __len__(self):
        """Get the length of the rectangle mesh.

        :return: The number of points in the rectangle mesh.
        """
        return len(self.mesh)

    def __getitem__(self, idx):
        """Get a specific point from the rectangle mesh using indexing.

        :param idx: The index of the desired point.
        :return: The point value at the specified index.
        """

        return self.mesh[idx, 0]


class RectangularPrism:
    """Initialize a 3D spatial domain."""

    def __init__(self, x_interval, y_interval, z_interval, shape):
        """Initialize a Rectangular Prism object to represent a three-dimensional shape.

        :param x_interval: A tuple or list representing the x-axis interval [start_x, end_x].
        :param y_interval: A tuple or list representing the y-axis interval [start_y, end_y].
        :param z_interval: A tuple or list representing the z-axis interval [start_z, end_z].
        :param shape: The number of points along each axis in the cube.
        """

        self.x_interval = x_interval
        self.y_interval = y_interval
        self.z_interval = z_interval
        self.shape = shape

    def generate_mesh(self, t_points):
        """Generate a mesh for the 3D spatial cube.

        :param t_points: The number of time points in the mesh.
        :return: A mesh for the 3D spatial cube.
        """

        x = np.linspace(self.x_interval[0], self.x_interval[1], num=self.shape[0])
        y = np.linspace(self.y_interval[0], self.y_interval[1], num=self.shape[1])
        z = np.linspace(self.z_interval[0], self.z_interval[1], num=self.shape[2])

        self.xx, self.yy = np.meshgrid(x, y)
        self.spatial_mesh = np.stack((self.xx.flatten(), self.yy.flatten()), 1)

        self.mesh = np.zeros((np.prod(self.shape), t_points, 3))

        for i in range(3):
            self.mesh[:, :, i] = np.tile(self.spatial_mesh[:, i], (t_points, 1)).T

        return self.mesh

    def __len__(self):
        """Get the length of the cube mesh.

        :return: The number of points in the cube mesh.
        """
        return len(self.mesh)

    def __getitem__(self, idx):
        """Get a specific point from the cube mesh using indexing.

        :param idx: The index of the desired point.
        :return: The point value at the specified index.
        """
        return self.mesh[idx, 0]



import numpy as np
import torch

from pinnstorch import utils

from .sampler_base import SamplerBase
from pinnstorch.utils import set_requires_grad

class DirichletBoundaryCondition(SamplerBase):
    """Initialize Dirichlet boundary condition."""

    def __init__(
        self,
        mesh,
        solution,
        num_sample: int = None,
        idx_t: int = None,
        boundary_fun=None,
        discrete: bool = False,
    ):
        """Initialize a mesh sampler for collecting training data in upper and lower boundaries for
        Dirichlet boundary condition.

        :param mesh: Instance of the mesh used for sampling.
        :param solution: Names of the solution outputs.
        :param num_sample: Number of samples to generate.
        :param idx_t: Index of the time step for discrete mode.
        :param boundary_fun: A function can apply on boundary data.
        :param discrete: It is a boolean that is true when problem is discrete.
        """

        super().__init__()

        self.solution_names = solution
        self.discrete = discrete

        spatial_upper_bound, time_upper_bound, solution_upper_bound = mesh.on_upper_boundary(
            self.solution_names
        )
        spatial_lower_bound, time_lower_bound, solution_lower_bound = mesh.on_lower_boundary(
            self.solution_names
        )

        spatial_bound = np.vstack([spatial_upper_bound, spatial_lower_bound])
        time_bound = np.vstack([time_upper_bound, time_lower_bound])

        solution_bound = {}
        for solution_name in self.solution_names:
            solution_bound[solution_name] = np.vstack(
                [solution_upper_bound[solution_name], solution_lower_bound[solution_name]]
            )

        if boundary_fun:
            solution_bound = boundary_fun(time_bound)

        self.idx_t = idx_t

        (
            self.spatial_domain_sampled,
            self.time_domain_sampled,
            self.solution_sampled,
        ) = self.sample_mesh(num_sample, (spatial_bound, time_bound, solution_bound))

        self.spatial_domain_sampled = list(torch.split(self.spatial_domain_sampled, (1), dim=1))
        self.solution_sampled = list(torch.split(self.solution_sampled, (1), dim=1))

    def sample_mesh(self, num_sample, flatten_mesh):
        """Sample the mesh data for training. If idx_t is defined, only points on that time will be
        selected. If num_sample is not defined the whole points will be selected.

        :param num_sample: Number of samples to generate.
        :param flatten_mesh: Flattened mesh data.
        :return: Sampled spatial, time, and solution data.
        """

        flatten_mesh = self.concatenate_solutions(flatten_mesh)

        if self.discrete:
            t_points = len(flatten_mesh[0]) // 2
            flatten_mesh = [
                np.vstack(
                    (
                        flatten_mesh_[self.idx_t : self.idx_t + 1, :],
                        flatten_mesh_[self.idx_t + t_points : self.idx_t + t_points + 1, :],
                    )
                )
                for flatten_mesh_ in flatten_mesh
            ]

        if num_sample is None:
            return self.convert_to_tensor(flatten_mesh)
        else:
            idx = np.random.choice(range(flatten_mesh[0].shape[0]), num_sample, replace=False)
            return self.convert_to_tensor(
                (flatten_mesh[0][idx, :], flatten_mesh[1][idx, :], flatten_mesh[2][idx, :])
            )

    def _loss_fn(self, inputs, loss):
        """Compute the loss function based on inputs and functions.

        :param inputs: Input data for computing the loss.
        :param loss: Loss variable.
        :return: Loss variable and outputs dict from the forward pass.
        """

        x, t, u = inputs

        # In discrete mode, we do not use time.
        if self.discrete:
            t = None

        outputs = self.functions["forward"](x, t)

        loss = self.functions["loss_fn"](loss, outputs, u, keys=self.solution_names)

        return loss, outputs


class PeriodicBoundaryCondition(SamplerBase):
    """Initialize Periodic boundary condition."""

    def __init__(
        self,
        mesh,
        solution,
        num_sample: int = None,
        idx_t: int = None,
        derivative_order: int = 0,
        discrete: bool = False,
    ):
        super().__init__()
        """Initialize a mesh sampler for collecting training data in upper and lower boundaries for
        periodic boundary condition.

        :param mesh: Instance of the mesh used for sampling.
        :param solution: Names of the solution outputs.
        :param num_sample: Number of samples to generate.
        :param idx_t: Index of the time step for discrete mode.
        :param derivative_order: Order of derivative.
        :param discrete: It is a boolean that is true when problem is discrete.
        """

        self.derivative_order = derivative_order
        self.idx_t = idx_t
        self.solution_names = solution

        spatial_upper_bound, time_upper_bound, _ = mesh.on_upper_boundary(self.solution_names)
        spatial_lower_bound, time_lower_bound, _ = mesh.on_lower_boundary(self.solution_names)

        self.discrete = discrete

        (self.spatial_domain_sampled, self.time_domain_sampled) = self.sample_mesh(
            num_sample,
            (spatial_upper_bound, time_upper_bound, spatial_lower_bound, time_lower_bound),
        )

        self.mid = len(self.time_domain_sampled) // 2
        self.spatial_domain_sampled = list(torch.split(self.spatial_domain_sampled, (1), dim=1))

    def sample_mesh(self, num_sample, flatten_mesh):
        """Sample the mesh data for training.

        :param num_sample: Number of samples to generate.
        :param flatten_mesh: Flattened mesh data.
        :return: Sampled spatial, time, and solution data.
        """

        if self.discrete:
            flatten_mesh = [
                flatten_mesh_[self.idx_t : self.idx_t + 1, :] for flatten_mesh_ in flatten_mesh
            ]

        if num_sample is None:
            return self.convert_to_tensor(
                (
                    np.vstack((flatten_mesh[0], flatten_mesh[2])),
                    np.vstack((flatten_mesh[1], flatten_mesh[3])),
                )
            )
        else:
            idx = np.random.choice(range(flatten_mesh[0].shape[0]), num_sample, replace=False)
            return self.convert_to_tensor(
                (
                    np.vstack((flatten_mesh[0][idx, :], flatten_mesh[2][idx, :])),
                    np.vstack((flatten_mesh[1][idx, :], flatten_mesh[3][idx, :])),
                )
            )
    
    def _loss_fn(self, inputs, loss):
        """Compute the loss function based on inputs and functions.

        :param inputs: Input data for computing the loss.
        :param loss: Loss variable.
        :return: Loss variable and outputs dict from the forward pass.
        """

        x, t, u = inputs

        # In discrete mode, we do not use time.
        if self.discrete:
            t = None

        outputs = self.functions["forward"](x, t)
        
        if self.derivative_order == 1:
            for solution_name in self.solution_names:
                if self.discrete:
                    outputs["tmp"] = utils.fwd_gradient(outputs[solution_name], x)[0]
                else:
                    outputs["tmp"] = utils.gradient(outputs[solution_name], x)[0]
                loss = self.functions["loss_fn"](loss, outputs, keys=["tmp"], mid=self.mid)

        loss = self.functions["loss_fn"](loss, outputs, keys=self.solution_names, mid=self.mid)

        return loss, outputs



from dataclasses import dataclass
from typing import Dict, List

import numpy as np


@dataclass
class PointCloudData:
    """A data class to hold point cloud data, including spatial points, time values, and solution
    data.

    This is an alternative class to spatial and time class that should be used along with `PointCloud` class.
    """

    spatial: List[np.array]  # List of spatial point data.
    time: np.array  # NumPy array containing time values.
    solution: Dict[str, np.array]  # Dictionary containing solution data.



import torch


class PINNDataLoader:
    """Custom DataLoader for the PINN datasets.

    This allows you to have a fast dataloader.
    """

    def __init__(self, dataset, batch_size=None, ignore=False, shuffle=False):
        """Initialize a PINNDataLoader.

        :param dataset: The dataset to load data from.
        :param batch_size: The batch size for the dataloader.
        :param ignore: Whether to ignore incomplete batches (default is False).
        :param shuffle: Whether to shuffle the dataset (default is False).
        """
        self.dataset = dataset
        self.batch_size = batch_size
        self.dataset_size = len(self.dataset)
        self.shuffle = shuffle
        self.current_index = 0
        self.ignore = ignore

        if self.shuffle:
            self.indices = torch.randperm(self.dataset_size)
        else:
            self.indices = torch.arange(self.dataset_size)

    def __len__(self):
        """Get the number of batches in the dataloader.

        :return: The number of batches.
        """
        if self.batch_size is None:
            return 1
        if self.ignore:
            return self.dataset_size // self.batch_size
        else:
            return (self.dataset_size // self.batch_size) + 1

    def __iter__(self):
        """Initialize the data loader iterator.

        :return: The data loader iterator.
        """
        self.current_index = 0
        return self

    def __next__(self):
        """Generate the next batch of data.

        :return: The next batch of data.
        """
        if self.current_index >= len(self.indices):
            raise StopIteration

        # If batch_size is None, return the entire dataset as a single batch
        if self.batch_size is None:
            batch = self.dataset[:]
            self.current_index += self.dataset_size
            return batch

        batch_indices = self.indices[self.current_index : self.current_index + self.batch_size]
        batch = self.dataset[batch_indices]
        self.current_index += self.batch_size

        return batch



from .dataloader.dataloader import PINNDataLoader
from .domains.point_cloud import PointCloudData
from .domains.spatial import Interval, Rectangle, RectangularPrism
from .domains.time import TimeDomain
from .mesh.mesh import Mesh, PointCloud
from .pinn_datamodule import PINNDataModule
from .sampler.boundary_condition import (
    DirichletBoundaryCondition,
    PeriodicBoundaryCondition,
)
from .sampler.initial_condition import InitialCondition
from .sampler.mesh_sampler import DiscreteMeshSampler, MeshSampler



#!/usr/bin/env python

from setuptools import setup, find_packages

# read the contents of your README file
from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name="pinnstorch",
    version="0.1.1",
    description="An implementation of PINNs in PyTorch using Lightning and Hydra.",
    author="Reza Akbarian Bafghi",
    author_email="reza.akbarianbafghi@coloardo.edu",
    long_description=long_description,
    long_description_content_type='text/markdown',
    url="https://github.com/rezaakb/pinns-torch",
    license='BSD-3-Clause',
    install_requires=["hydra-core", "scipy", "pyDOE", "matplotlib", "rootutils", "tqdm", "rich"],
    packages=find_packages(include='pinnstorch.*'),
    # use this to customize global commands available in the terminal after installing the package
    include_package_data=True,
    classifiers=[
    'Development Status :: 2 - Pre-Alpha',
    'Programming Language :: Python :: 3.9',
    'Programming Language :: Python :: 3.10',
],
)



from typing import Callable, List, Union

import numpy as np
import torch
from pyDOE import lhs

from pinnstorch.data import Interval, Rectangle, RectangularPrism, TimeDomain


class MeshBase:
    """This helper class is utilized by the Mesh and PointCloud classes."""

    def __init__():
        """Base class for generating mesh data and boundary conditions."""

        pass

    def domain_bounds(self):
        """Calculate the domain bounds based on the generated spatial and time domain mesh.

        :return: Lower and upper bounds of the domain.
        """
        mesh = np.hstack(
            (
                self.spatial_domain_mesh.reshape(-1, self.spatial_dim),
                self.time_domain_mesh.reshape(-1, 1),
            )
        )

        ub = mesh.max(0)
        lb = mesh.min(0)
        return lb, ub

    def on_lower_boundary(self, solution_names: List):
        """Generate data for points on the lower boundary.

        :param solution_names: Names of the solution outputs.
        :return: Spatial, time, and solution data on the lower boundary.
        """
        spatial_domain = (
            np.ones((self.time_domain_mesh.shape[1], self.spatial_dim)) * self.lb[0:-1]
        )
        time_domain = self.time_domain_mesh[0, :]
        solution_domain = {
            solution_name: self.solution[solution_name][0, :][:, None]
            for solution_name in solution_names
        }

        return spatial_domain, time_domain, solution_domain

    def on_upper_boundary(self, solution_names: List):
        """Generate data for points on the upper boundary.

        :param solution_names: Names of the solution outputs.
        :return: Spatial, time, and solution data on the upper boundary.
        """
        spatial_domain = (
            np.ones((self.time_domain_mesh.shape[1], self.spatial_dim)) * self.ub[0:-1]
        )
        time_domain = self.time_domain_mesh[-1, :]
        solution_domain = {
            solution_name: self.solution[solution_name][-1, :][:, None]
            for solution_name in solution_names
        }
        return spatial_domain, time_domain, solution_domain

    def on_initial_boundary(self, solution_names: List, idx: int = 0):
        """Generate data for points on the initial boundary.

        :param solution_names: Names of the solution outputs.
        :param idx: Index of the time step.
        :return: Spatial, time, and solution data on the initial boundary.
        """

        spatial_domain = np.squeeze(self.spatial_domain_mesh[:, idx : idx + 1, :], axis=-2)
        time_domain = self.time_domain_mesh[:, idx]
        solution_domain = {
            solution_name: self.solution[solution_name][:, idx : idx + 1]
            for solution_name in solution_names
        }

        return spatial_domain, time_domain, solution_domain

    def collection_points(self, N_f: int, use_lhs: bool = True):
        """Generate a collection of points for data collection.

        :param N_f: Number of points to collect.
        :return: Collection of points in the spatial domain.
        """
        if use_lhs:
            f = self.lb + (self.ub - self.lb) * lhs(self.spatial_dim + 1, N_f)
            spatial_domain = f[:, 0 : self.spatial_dim]
            time_domain = f[:, self.spatial_dim : self.spatial_dim + 1]
        else:
            spatial_domain, time_domain, _ = self.flatten_mesh(None)
        return spatial_domain, time_domain

    def flatten_mesh(self, solution_names: List):
        """Flatten the mesh data for training.

        :param solution_names: Names of the solution outputs.
        :return: Flattened spatial, time, and solution data.
        """
        time_domain = self.time_domain_mesh.flatten()[:, None]
        spatial_domain = np.zeros((len(time_domain), self.spatial_domain_mesh.shape[-1]))
        for i in range(self.spatial_domain_mesh.shape[-1]):
            spatial_domain[:, i] = self.spatial_domain_mesh[:, :, i].flatten()

        solution_domain = {}
        if solution_names is not None:
            for solution_name in solution_names:
                solution_domain[solution_name] = self.solution[solution_name][:, :].flatten()[
                    :, None
                ]

        return spatial_domain, time_domain, solution_domain


class Mesh(MeshBase):
    """For using this class you should define a SpatialDomain and TimeDomain classes.

    If dimensions of mesh is not determined, it is better to use PointCloud.
    """

    def __init__(
        self,
        spatial_domain: Union[Interval, Rectangle, RectangularPrism],
        time_domain: TimeDomain,
        root_dir: str,
        read_data_fn: Callable,
        ub: List = None,
        lb: List = None,
    ):
        """Generate a mesh based on spatial and time domains, and load solution data.

        :param spatial_domain: Instance of a SpatialDomain class.
        :param time_domain: Instance of a TimeDomain class.
        :param root_dir: Root directory for solution data.
        :param read_data_fn: Function to read solution data.
        :param ub: Upper bounds for domain.
        :param lb: Lower bounds for domain.
        """

        self.solution = read_data_fn(root_dir)
        spatial_points, t_points = list(self.solution.values())[0].shape

        self.spatial_domain, self.time_domain = spatial_domain, time_domain

        # Generate Mesh for both spatial and time domain
        self.spatial_domain_mesh = spatial_domain.generate_mesh(t_points)
        self.time_domain_mesh = time_domain.generate_mesh(spatial_points)

        self.spatial_dim = self.spatial_domain_mesh.shape[-1]

        # We utilize the user-defined lower and upper bounds if they are specified.
        if ub is None and lb is None:
            self.lb, self.ub = self.domain_bounds()
        else:
            self.lb, self.ub = np.array(lb), np.array(ub)


class PointCloud(MeshBase):
    """For using this class you should define a mesh of spatial domain, time domain, and
    solutions."""

    def __init__(self, root_dir: str, read_data_fn: Callable, ub: List = None, lb: List = None):
        """Generate a point cloud mesh and load data from files.

        :param root_dir: Root directory for data.
        :param read_data_fn: Function to read spatial, time, and solution data.
        :param ub: Upper bounds for domain.
        :param lb: Lower bounds for domain.
        """

        data = read_data_fn(root_dir)
        self.spatial_domain, self.time_domain, self.solution = (
            data.spatial,
            data.time,
            data.solution,
        )

        if not isinstance(self.solution, dict):
            raise "Solution outputs of read_data_fn function is not dictionary."

        if isinstance(self.time_domain, list):
            if len(self.time_domain) == 1:
                self.time_domain = self.time_domain[0]

        if not isinstance(self.spatial_domain, list):
            self.spatial_domain = [self.spatial_domain]

        spatial_num_points, time_num_points = list(self.solution.values())[0].shape

        self.spatial_dim = len(self.spatial_domain)
        self.time_dim = 1
        self.solution_dim = len(self.solution.keys())

        # Generate Mesh for both spatial and time domain
        self.spatial_domain_mesh = np.zeros(
            (spatial_num_points, time_num_points, self.spatial_dim)
        )

        for i, interval in enumerate(self.spatial_domain):
            self.spatial_domain_mesh[:, :, i] = np.tile(interval, (1, time_num_points))

        self.time_domain_mesh = np.tile(self.time_domain, (1, spatial_num_points)).T[:, :, None]

        # We utilize the user-defined lower and upper bounds if they are specified.
        if ub is None and lb is None:
            self.lb, self.ub = self.domain_bounds()
        else:
            self.lb, self.ub = np.array(lb), np.array(ub)
